"""This module implements the EmbedAllPermutationsPass pass."""
from __future__ import annotations

import copy
import itertools as it
import logging
from typing import Callable

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.machine import MachineModel
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.passes.mapping.topology import SubtopologySelectionPass
from bqskit.passes.synthesis.leap import LEAPSynthesisPass
from bqskit.passes.synthesis.synthesis import SynthesisPass
from bqskit.qis.graph import CouplingGraph
from bqskit.qis.permutation import PermutationMatrix
from bqskit.runtime import get_runtime


_logger = logging.getLogger(__name__)


def multi_qudit_op_count(circuit: Circuit) -> float:
    """Counts the number of multi-qudit operations in a circuit."""
    x = sum(c for g, c in circuit.gate_counts.items() if g.num_qudits >= 2)
    return float(x)


class EmbedAllPermutationsPass(BasePass):
    """Embed permutation aware synthesis results into a flow for future use."""

    def __init__(
        self,
        input_perm: bool = False,
        output_perm: bool = True,
        vary_topology: bool = True,
        inner_synthesis: SynthesisPass = LEAPSynthesisPass(),
        scoring_fn: Callable[[Circuit], float] = multi_qudit_op_count,
    ) -> None:
        """
        Construct a EmbedAllPermutationsPass.

        Args:
            input_perm (bool): If true, vary the input permutation
                during synthesis. (Default: False)

            output_perm (bool): If true, vary the output permutation
                during synthesis. (Default: True)

            vary_topology (bool): If true, vary the desired coupling graph
                during synthesis. (Default: True)

            inner_synthesis (SynthesisPass): The synthesis algorithm used
                on all permutations. (Default: :class:`LEAPSynthesisPass`)

            scoring_fn (Callable[[Circuit], float]): The scoring function
                used when comparing the original circuit to the synthesized
                circuit with the same configuration. The smallest score wins.
                (Default: :func:`multi_qudit_op_count`)
        """
        if not isinstance(inner_synthesis, SynthesisPass):
            bad_type = type(inner_synthesis)
            raise TypeError(f'Expected SynthesisPass object, got {bad_type}.')

        if not callable(scoring_fn):
            bad_type = type(scoring_fn)
            m = f'Expected a function from circuits to scores, got {bad_type}.'
            raise TypeError(m)

        self.input_perm = input_perm
        self.output_perm = output_perm
        self.vary_topology = vary_topology
        self.inner_synthesis = inner_synthesis
        self.scoring_fn = scoring_fn

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        utry = data.target

        if not all(r == utry.radixes[0] for r in utry.radixes):
            raise NotImplementedError(
                'PermutationAwareSynthesisPass only supports unitaries '
                'with the same radix on all qudits currently.',
            )

        # Calculate all permuted targets
        width = utry.num_qudits
        perms = list(it.permutations(range(width)))
        no_perm = [tuple(range(width))]
        Pis = [
            PermutationMatrix.from_qudit_location(width, utry.radixes[0], p)
            for p in perms
        ]
        Pos = [
            PermutationMatrix.from_qudit_location(width, utry.radixes[0], p)
            for p in perms
        ]

        if self.input_perm and self.output_perm:
            permsbyperms = list(it.product(perms, perms))
            targets = [Po.T @ utry @ Pi for Pi, Po in it.product(Pis, Pos)]

        elif self.input_perm:
            permsbyperms = list(it.product(perms, no_perm))
            targets = [utry @ Pi for Pi in Pis]

        elif self.output_perm:
            permsbyperms = list(it.product(no_perm, perms))
            targets = [Po.T @ utry for Po in Pos]

        else:
            _logger.warning('No permutation is being used in PAS.')
            permsbyperms = list(it.product(no_perm, no_perm))
            targets = [utry]

        # Calculate all target coupling graphs
        if self.vary_topology and width != 1:
            if SubtopologySelectionPass.key not in data:
                raise RuntimeError(
                    'Cannot find subtopologies, try running a'
                    ' SubtopologySelectionPass first.',
                )

            if width not in data[SubtopologySelectionPass.key]:
                raise RuntimeError(
                    'Subtopology information for block size'
                    f' {width} is not available.',
                )

            graphs = data[SubtopologySelectionPass.key][width]

        else:
            graphs = [CouplingGraph.all_to_all(width)]

        datas = []
        for graph in graphs:
            model = MachineModel(
                circuit.num_qudits, graph,
                data.gate_set, data.model.radixes,
            )
            target_data = copy.deepcopy(data)
            target_data.model = model
            datas.append(target_data)

        # Create parallel arrays for map
        extended_targets = []
        extended_datas = []
        for t, d in it.product(targets, datas):
            extended_targets.append(t)
            extended_datas.append(d)

        # Synthesize all permuted targets
        circuits: list[Circuit] = await get_runtime().map(
            self.inner_synthesis.synthesize,
            extended_targets,
            extended_datas,
        )

        # Store results
        perm_data: dict[
            CouplingGraph,
            dict[tuple[tuple[int, ...], tuple[int, ...]], Circuit],
        ] = {}
        for i, c in enumerate(circuits):
            graph = graphs[i % len(graphs)]
            perm = permsbyperms[i // len(graphs)]
            if graph not in perm_data:
                perm_data[graph] = {}

            if perm in perm_data[graph]:
                # Update if it is better than whats already there
                s1 = self.scoring_fn(perm_data[graph][perm])
                s2 = self.scoring_fn(c)
                if s2 < s1:
                    perm_data[graph][perm] = c
            else:
                perm_data[graph][perm] = c

            # Calculate number of multi-qudit gates
            num_mq_gates = 0
            for gate, count in c.gate_counts.items():
                if gate.num_qudits >= 2:
                    num_mq_gates += count

            # Generate the extra circuits through universal permutations
            all_perms = list(it.permutations(range(width)))
            for univ_perm in all_perms[1:]:
                renumber_c = c.copy()
                renumber_c.renumber_qudits(univ_perm)
                new_pi = tuple(univ_perm[i] for i in perm[0])
                new_pf = tuple(univ_perm[i] for i in perm[1])
                new_graph = renumber_c.coupling_graph
                if new_graph not in perm_data:
                    perm_data[new_graph] = {}

                new_perm = (new_pi, new_pf)
                if new_perm not in perm_data[new_graph]:
                    perm_data[new_graph][new_perm] = renumber_c
                else:
                    s1 = self.scoring_fn(perm_data[new_graph][new_perm])
                    s2 = self.scoring_fn(renumber_c)
                    if s2 < s1:
                        perm_data[new_graph][new_perm] = renumber_c

        # Override no perm result if original is better and compatible
        if circuit.gate_set.issubset(data.model.gate_set):
            for univ_perm in it.permutations(range(width)):
                # Permute original circuit and override worse results
                uperm = (univ_perm, univ_perm)
                renumber_c = circuit.copy()
                renumber_c.renumber_qudits(univ_perm)
                new_graph = renumber_c.coupling_graph
                new_score = self.scoring_fn(renumber_c)
                for graph, graph_data in perm_data.items():
                    if all(e in graph for e in new_graph):
                        if uperm not in graph_data:
                            graph_data[uperm] = renumber_c
                        else:
                            if new_score < self.scoring_fn(graph_data[uperm]):
                                graph_data[uperm] = renumber_c

        # Record permutation data in the pass data
        data['permutation_data'] = perm_data



"""This module implements the Instantiater base class."""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

from bqskit.ir.opt.cost.functions import HilbertSchmidtCostGenerator
from bqskit.ir.opt.multistartgens.random import RandomStartGenerator
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit
    from bqskit.qis.state.state import StateLike
    from bqskit.qis.state.system import StateSystemLike
    from bqskit.qis.unitary.unitarymatrix import UnitaryLike


class Instantiater(abc.ABC):
    """
    The Instantiater class.

    An Instantiater is responsible for instantiating circuit templates such that
    the resulting circuit bests implements the desired target.
    """

    @abc.abstractmethod
    def instantiate(
        self,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector | StateSystem,
        x0: npt.NDArray[np.float64],
    ) -> npt.NDArray[np.float64]:
        """
        Instantiate `circuit` to best implement `target`.

        Args:
            circuit (Circuit): The circuit template to instantiate.

            target (UnitaryMatrix | StateVector | StateSystem): The unitary
                matrix to implement or state to prepare.

            x0 (np.ndarray): Initial point to use during instantiation.

        Returns:
            (np.ndarray): The paremeters for the circuit that makes the
                circuit best implement `target`.

        Notes:
            This method should be side-effect free. This is necessary since
            many instantiate calls to the same circuit using the same
            Instantiater object may happen in parallel.
        """

    def multi_start_instantiate(
        self,
        circuit: Circuit,
        target: UnitaryLike | StateLike | StateSystemLike,
        num_starts: int,
    ) -> Circuit:
        """
        Instantiate `circuit` to best implement `target` with multiple starts.

        Args:
            circuit (Circuit): The circuit template to instantiate.

            target (UnitaryMatrix | StateVector | StateSystem): The unitary
                matrix to implement or state to prepare.

            num_starts (int): The number of starting points to attempt
                instantiation with.

        Returns:
            (Circuit): A circuit copy with the best parameters with
                respect to `target`.

        Notes:
            This method should be side-effect free. This is necessary since
            many instantiate calls to the same circuit using the same
            Instantiater object may happen in parallel.
        """
        _circuit = circuit.copy()
        self.multi_start_instantiate_inplace(_circuit, target, num_starts)
        return _circuit

    def multi_start_instantiate_inplace(
        self,
        circuit: Circuit,
        target: UnitaryLike | StateLike | StateSystemLike,
        num_starts: int,
    ) -> None:
        """
        Instantiate `circuit` to best implement `target` with multiple starts.

        See :func:`multi_start_instantiate` for more info.

        Notes:
            This method is a version of :func:`multi_start_instantiate`
            that modifies `circuit` in place rather than returning a copy.
        """
        target = self.check_target(target)
        start_gen = RandomStartGenerator()
        starts = start_gen.gen_starting_points(num_starts, circuit, target)
        cost_fn = HilbertSchmidtCostGenerator().gen_cost(circuit, target)
        params_list = [self.instantiate(circuit, target, x0) for x0 in starts]
        params = sorted(params_list, key=lambda x: cost_fn(x))[0]
        circuit.set_params(params)

    async def multi_start_instantiate_async(
        self,
        circuit: Circuit,
        target: UnitaryLike | StateLike | StateSystemLike,
        num_starts: int,
    ) -> Circuit:
        """
        Instantiate `circuit` to best implement `target` with multiple starts.

        See :func:`multi_start_instantiate` for more info.

        Notes:
            This method is an async version of :func:`multi_start_instantiate`
            and designed to parallelize the instantiation calls in the
            BQSKit Runtime during pass execution.
        """
        from bqskit.runtime import get_runtime
        target = self.check_target(target)
        start_gen = RandomStartGenerator()
        starts = start_gen.gen_starting_points(num_starts, circuit, target)
        cost_fn = HilbertSchmidtCostGenerator().gen_cost(circuit, target)
        params_list = await get_runtime().map(
            self.instantiate,
            [circuit] * num_starts,
            [target] * num_starts,
            starts,
        )
        params = sorted(params_list, key=lambda x: cost_fn(x))[0]
        circuit.set_params(params)
        return circuit

    @staticmethod
    @abc.abstractmethod
    def is_capable(circuit: Circuit) -> bool:
        """Return true if the circuit can be instantiated."""

    @staticmethod
    @abc.abstractmethod
    def get_violation_report(circuit: Circuit) -> str:
        """
        Return a message explaining why `circuit` cannot be instantiated.

        Args:
            circuit (Circuit): Generate a report for this circuit.

        Raises:
            ValueError: If `circuit` can be instantiated with this
                instantiater.
        """

    def check_target(
        self,
        target: UnitaryLike | StateLike | StateSystemLike,
    ) -> UnitaryMatrix | StateVector | StateSystem:
        """Check `target` to be valid and return it casted."""
        # Check `target`
        try:
            if UnitaryMatrix.is_unitary(target):
                target = UnitaryMatrix(target)

            elif StateVector.is_pure_state(target):
                target = StateVector(target)

            elif StateSystem.is_state_system(target):
                target = StateSystem(target)

            else:
                raise TypeError(
                    'Target is neither a unitary, a state system'
                    f', nor a state. Got {type(input)}.',
                )
        except Exception as e:
            raise TypeError(
                'Unable to determine type of input.'
                ' Ensure that you are trying to compile a valid'
                ' circuit, unitary, or state.',
            ) from e

        return target

    @staticmethod
    @abc.abstractmethod
    def get_method_name() -> str:
        """Return the name of this method."""



"""This module implements the CircuitStructure Class."""
from __future__ import annotations

from bqskit.ir.circuit import Circuit


class CircuitStructure:
    """Stores compressed positions of gates in a circuit as a hashable type."""

    def __init__(self, circuit: Circuit) -> None:
        """
        Construct a CircuitStructure object.

        Args:
            circuit (Circuit): The circuit to store the structure of.
        """

        if not isinstance(circuit, Circuit):
            raise TypeError(f'Expected a circuit, got {type(circuit)}.')

        structure = []

        # Compress circuit without modifying input circuit
        compressed_circuit = Circuit(circuit.num_qudits, circuit.radixes)
        for op in circuit:
            compressed_circuit.append(op)

        # Go through gates and add gate names and locations to structure array
        current_cycle = ['No Gate'] * compressed_circuit.num_qudits
        current_cycle_index = 0

        for cycle, op in compressed_circuit.operations_with_cycles():
            if cycle != current_cycle_index:
                structure.append(tuple(current_cycle))
                current_cycle = ['No Gate'] * compressed_circuit.num_qudits
                current_cycle_index = cycle

            for qudit in op.location:
                current_cycle[qudit] = str(op)

        structure.append(tuple(current_cycle))
        self.structure = tuple(structure)
        self.hash = hash(self.structure)

    def __hash__(self) -> int:
        """Get the hash of the circuit structure."""
        return self.hash

    def __eq__(self, other: object) -> bool:
        """Return true if `self` has the same structure as `other`."""
        return (
            isinstance(other, CircuitStructure)
            and self.structure == other.structure
        )



"""This module implements the SinglePhysicalPredicate class."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from bqskit.ir.gates.constant.sx import SXGate
from bqskit.ir.gates.generalgate import GeneralGate
from bqskit.ir.gates.parameterized.rx import RXGate
from bqskit.ir.gates.parameterized.rz import RZGate
from bqskit.ir.gates.parameterized.u1 import U1Gate
from bqskit.passes.control.predicate import PassPredicate

if TYPE_CHECKING:
    from bqskit.compiler.passdata import PassData
    from bqskit.ir.circuit import Circuit

_logger = logging.getLogger(__name__)


class SinglePhysicalPredicate(PassPredicate):
    """
    The SinglePhysicalPredicate class.

    The SinglePhysicalPredicate returns true if circuit's single-qudit gates are
    in the native gate set.
    """

    def get_truth_value(self, circuit: Circuit, data: PassData) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""
        for gate in circuit.gate_set:
            if gate.num_qudits > 1:
                continue
            if gate not in data.gate_set:
                _logger.debug(f'{gate} not in native set: {data.gate_set}.')
                return False

        return True


class NoSingleQuditGatesInModel(PassPredicate):
    """A predicate that checks if the model has single qudit gates."""

    def get_truth_value(self, circuit: Circuit, data: PassData) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""
        return len(data.gate_set.single_qudit_gates) == 0


class HasGeneralSingleQuditGate(PassPredicate):
    """A predicate that checks if the model has a general single qudit gate."""

    def get_truth_value(self, circuit: Circuit, data: PassData) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""
        return any(
            isinstance(g, GeneralGate)
            for g in data.gate_set.single_qudit_gates
        )


class ZXGatePredicate(PassPredicate):
    """A predicate that checks if the model has a RZ or U1 and SX or RX gate."""

    def get_truth_value(self, circuit: Circuit, data: PassData) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""
        return (
            (
                RZGate() in data.gate_set.single_qudit_gates
                or U1Gate() in data.gate_set.single_qudit_gates
            ) and (
                SXGate() in data.gate_set.single_qudit_gates
                or RXGate() in data.gate_set.single_qudit_gates
            )
        )


class AllConstantSingleQuditGates(PassPredicate):
    """A predicate that checks if all single qudit gates are constant."""

    def get_truth_value(self, circuit: Circuit, data: PassData) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""
        return all(g.is_constant() for g in data.gate_set.single_qudit_gates)



"""This package contains classes and functions necessary to test BQSKit."""
from __future__ import annotations

from bqskit.utils.test.strategies import circuit_regions
from bqskit.utils.test.strategies import cycle_intervals
from bqskit.utils.test.strategies import everything_except

__all__ = [
    'circuit_regions',
    'cycle_intervals',
    'everything_except',
]



"""This module implements the GateSet class."""
from __future__ import annotations

import sys
from collections.abc import Set
from typing import Any
from typing import Iterable
from typing import Iterator
from typing import Sequence
from typing import TYPE_CHECKING
from typing import Union

from bqskit.ir.gate import Gate
from bqskit.ir.gates.constant.csum import CSUMGate
from bqskit.ir.gates.constant.cx import CNOTGate
from bqskit.ir.gates.generalgate import GeneralGate
from bqskit.ir.gates.parameterized.cphase import ArbitraryCPhaseGate
from bqskit.ir.gates.parameterized.u3 import U3Gate
from bqskit.ir.gates.parameterized.u8 import U8Gate
from bqskit.ir.gates.parameterized.unitary import VariableUnitaryGate
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_sequence_of_int

if TYPE_CHECKING:
    from bqskit.passes.search.generator import LayerGenerator

if sys.version_info < (3, 9):
    SuperType = Set
else:
    SuperType = Set[Gate]


class GateSet(SuperType):
    """A set of a quantum processing unit's supported quantum gates."""

    def __init__(self, gates: GateSetLike) -> None:
        """Initialize a GateSet object."""
        if isinstance(gates, GateSet):
            self._gates: frozenset[Gate] = gates._gates.copy()
            self.radix_set: frozenset[int] = gates.radix_set.copy()
            return

        if isinstance(gates, Gate):
            gates = [gates]

        radix_set: set[int] = set()
        gate_set: set[Gate] = set()
        for g in gates:
            radix_set.update(g.radixes)
            gate_set.add(g)
        self.radix_set = frozenset(radix_set)
        self._gates = frozenset(gate_set)

    def build_layer_generator(self) -> LayerGenerator:
        """Build a standard layer generator compliant with this gate set."""
        if not any(isinstance(g, GeneralGate) for g in self.single_qudit_gates):
            raise RuntimeError(
                'Cannot automatically build a layer generator for a gate set'
                " without any general single-qubit gates. Instead, it's best"
                ' to use `build_mq_layer_generator` and then do single-qudit'
                ' gate retargeting afterwards.',
            )

        from bqskit.passes.search.generators import SimpleLayerGenerator
        from bqskit.passes.search.generators import WideLayerGenerator

        sqs = [g for g in self.single_qudit_gates if isinstance(g, GeneralGate)]
        sq = sqs[0]

        if len(self.two_qudit_gates) == 1 and len(self.many_qudit_gates) == 0:
            gate = list(self.two_qudit_gates)[0]
            return SimpleLayerGenerator(gate, sq)

        return WideLayerGenerator(self.multi_qudit_gates, sq)

    def build_mq_layer_generator(self) -> LayerGenerator:
        """
        Produce a layer generator using this gate set's multi-qudit gates.

        The resulting layer generator will use arbitrary single-qudit gates.
        """
        from bqskit.passes.search.generators import FourParamGenerator
        from bqskit.passes.search.generators import SimpleLayerGenerator
        from bqskit.passes.search.generators import WideLayerGenerator

        if len(self.two_qudit_gates) == 1 and len(self.many_qudit_gates) == 0:
            if CNOTGate() in self:
                return FourParamGenerator()

            gate = list(self.two_qudit_gates)[0]

            if isinstance(gate, GeneralGate):
                return WideLayerGenerator(gate, None)

            return SimpleLayerGenerator(gate, self.get_general_sq_gate())

        return WideLayerGenerator(
            self.multi_qudit_gates,
            self.get_general_sq_gate(),
        )

    def get_general_sq_gate(self) -> GeneralGate:
        """
        Return the gate set's arbitrary single-qudit gate.

        During off-the-shelf BQSKit compilations, single-qudit gate rebasing is
        done last. The earlier stages of compilation focus on multi-qudit gates
        and use arbitrary single-qudit rotations. This function will return the
        best arbitrary single-qudit rotation to use during these steps based on
        the gates in the set.
        """
        if len(self.radix_set) == 0:
            raise RuntimeError('Cannot suggest single-qudit gate for empty set')

        if len(self.radix_set) != 1:
            raise RuntimeError(
                'Cannot automatically suggest general single-qudit gate for'
                ' mixed radix systems.',
            )

        if U3Gate() in self:
            return U3Gate()

        if U8Gate() in self:
            return U8Gate()

        for g in self:
            if isinstance(g, VariableUnitaryGate) and g.num_qudits == 1:
                return g

        for g in self:
            if isinstance(g, GeneralGate) and g.num_qudits == 1:
                return g

        radix = list(self.radix_set)[0]

        if radix == 2:
            return U3Gate()

        return VariableUnitaryGate(1, [radix])

    @property
    def single_qudit_gates(self) -> set[Gate]:
        """All gates that act on only one qudit."""
        return {g for g in self if g.num_qudits == 1}

    @property
    def two_qudit_gates(self) -> set[Gate]:
        """All gates that act on two qudits."""
        return {g for g in self if g.num_qudits == 2}

    @property
    def many_qudit_gates(self) -> set[Gate]:
        """All gates that act on more than two qudits."""
        return {g for g in self if g.num_qudits > 2}

    @property
    def multi_qudit_gates(self) -> set[Gate]:
        """All gates that act on more than one qudits."""
        return {g for g in self if g.num_qudits >= 2}

    @staticmethod
    def default_gate_set(radixes: Sequence[int] | int = 2) -> GateSet:
        """Build a default gate set for a given `radixes`."""
        if is_integer(radixes):
            radixes = [radixes]

        assert is_sequence_of_int(radixes)

        if len(radixes) == 0 or all(r == 2 for r in radixes):
            return GateSet({CNOTGate(), U3Gate()})

        unique_radixes = set(radixes)

        gates: set[Gate] = set()
        for r in unique_radixes:
            gates.add(VariableUnitaryGate(1, [r]))

        if len(unique_radixes) == 1:
            gates.add(CSUMGate(list(unique_radixes)[0]))
            return GateSet(gates)

        for r1 in unique_radixes:
            for r2 in unique_radixes:
                if r1 <= r2:
                    gates.add(ArbitraryCPhaseGate([r1, r2]))

        return GateSet(gates)

    def __iter__(self) -> Iterator[Gate]:
        """Iterator for this gate set's gates."""
        return self._gates.__iter__()

    def __len__(self) -> int:
        """Number of gates in the set."""
        return len(self._gates)

    def __contains__(self, obj: object) -> bool:
        """Return true if a gate is in the set."""
        return self._gates.__contains__(obj)

    def union(self, *others: Any) -> GateSet:
        """Return a new GateSet with elements from this one and all others."""
        return GateSet(self._gates.union(*others))

    def intersection(self, *others: Any) -> GateSet:
        """Return a new GateSet with elements common to all sets."""
        return GateSet(self._gates.intersection(*others))

    def difference(self, *others: Any) -> GateSet:
        """Return a new GateSet with this set's elements but not the others."""
        return GateSet(self._gates.difference(*others))

    def symmetric_difference(self, other: Iterable[Gate]) -> GateSet:
        """Return a new GateSet with elements in either set but not both."""
        return GateSet(self._gates.symmetric_difference(other))

    def issubset(self, other: Iterable[Gate]) -> bool:
        """Test whether every element in the gate set is in other."""
        return self._gates.issubset(other)

    def issuperset(self, other: Iterable[Gate]) -> bool:
        """Report whether this gate set contains `other`."""
        return self._gates.issuperset(other)

    def __str__(self) -> str:
        """String representation of the GateSet."""
        return self._gates.__str__().replace('frozenset', 'GateSet')

    def __repr__(self) -> str:
        """Detailed representation of the GateSet."""
        return self._gates.__repr__().replace('frozenset', 'GateSet')


GateSetLike = Union[GateSet, Iterable[Gate], Gate]



"""
================================================
Compiler Infrastructure (:mod:`bqskit.compiler`)
================================================

The `bqskit.compiler` package implements the BQSKit's compiler framework.
This includes classes for defining, combining, executing, and managing
compilation algorithms. It also includes the standard :func:`compile`
function definition.

.. rubric:: Standard BQSKit Compile Function

.. autosummary::
    :toctree: autogen
    :recursive:
    :nosignatures:

    compile

.. rubric:: Compiler Infrastructure

.. autosummary::
    :toctree: autogen
    :recursive:
    :nosignatures:

    BasePass
    Compiler
    MachineModel
    GateSet
    GateSetLike
    PassData
    CompilationStatus
    CompilationTask
    Workflow
    WorkflowLike
"""
from __future__ import annotations

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.compile import compile
from bqskit.compiler.compiler import Compiler
from bqskit.compiler.gateset import GateSet
from bqskit.compiler.gateset import GateSetLike
from bqskit.compiler.machine import MachineModel
from bqskit.compiler.passdata import PassData
from bqskit.compiler.status import CompilationStatus
from bqskit.compiler.task import CompilationTask
from bqskit.compiler.workflow import Workflow
from bqskit.compiler.workflow import WorkflowLike

__all__ = [
    'BasePass',
    'compile',
    'Compiler',
    'GateSet',
    'GateSetLike',
    'MachineModel',
    'PassData',
    'CompilationStatus',
    'CompilationTask',
    'Workflow',
    'WorkflowLike',
]



"""
This module implements circuit metrics used in the Supermarq benchmark suite.

References:
    Tomesh, Teague, et al. "Supermarq: A scalable quantum benchmark suite."
    2022 IEEE International Symposium on High-Performance Computer
    Architecture (HPCA). IEEE, 2022.
"""
from __future__ import annotations

import numpy as np

from bqskit.ir.circuit import Circuit


def supermarq_program_communication(circuit: Circuit) -> float:
    """SupermarQ program communication metric."""
    n = circuit.num_qudits
    degrees_sum = sum(circuit.coupling_graph.get_qudit_degrees())
    return degrees_sum / (n * (n - 1))


def supermarq_critical_depth(circuit: Circuit) -> float:
    """SupermarQ critical depth metric."""
    qudit_depths = np.zeros(circuit.num_qudits, dtype=int)
    num_multi_qubit_gates = 0
    for op in circuit:
        if len(op.location) > 1:
            new_depth = max(qudit_depths[list(op.location)]) + 1
            qudit_depths[list(op.location)] = new_depth
            num_multi_qubit_gates += 1
    return int(max(qudit_depths)) / num_multi_qubit_gates


def supermarq_entanglement_ratio(circuit: Circuit) -> float:
    """SupermarQ entanglement-ratio metric."""
    num_multi_qubit_gates = 0
    num_gates = 0
    for op in circuit:
        num_gates += 1
        if len(op.location) > 1:
            num_multi_qubit_gates += 1
    return num_multi_qubit_gates / num_gates


def supermarq_parallelism(circuit: Circuit) -> float:
    """SupermarQ parallelism metric."""
    d = circuit.depth
    ng = circuit.num_operations
    return (ng / d - 1) / (circuit.num_qudits - 1)


def supermarq_liveness(circuit: Circuit) -> float:
    """SupermarQ liveness metric."""
    liveness_sum = 0
    for i in range(circuit.num_qudits):
        for j in range(circuit.num_cycles):
            if circuit._circuit[j][i] is not None:
                liveness_sum += 1
    return liveness_sum / (circuit.num_qudits * circuit.depth)



"""This package implements utility passes."""
from __future__ import annotations

from bqskit.passes.util.compress import CompressPass
from bqskit.passes.util.conversion import BlockConversionPass
from bqskit.passes.util.converttou3 import ToU3Pass
from bqskit.passes.util.converttovar import ToVariablePass
from bqskit.passes.util.extend import ExtendBlockSizePass
from bqskit.passes.util.fill import FillSingleQuditGatesPass
from bqskit.passes.util.log import LogErrorPass
from bqskit.passes.util.log import LogPass
from bqskit.passes.util.random import SetRandomSeedPass
from bqskit.passes.util.record import RecordStatsPass
from bqskit.passes.util.structure import StructureAnalysisPass
from bqskit.passes.util.unfold import UnfoldPass
from bqskit.passes.util.update import UpdateDataPass

__all__ = [
    'CompressPass',
    'RecordStatsPass',
    'SetRandomSeedPass',
    'UnfoldPass',
    'UpdateDataPass',
    'ToU3Pass',
    'BlockConversionPass',
    'LogPass',
    'ExtendBlockSizePass',
    'LogErrorPass',
    'FillSingleQuditGatesPass',
    'ToVariablePass',
    'StructureAnalysisPass',
]



"""This module implements the HGate."""
from __future__ import annotations

import numpy as np

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.quditgate import QuditGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer


class HGate(ConstantGate, QuditGate):
    """
    The one-qudit Hadamard gate. This is a Clifford gate.

    By default, the HGate is a qubit Hadamard gate. However, it can be
    generalized to qudits by setting the radix parameter during construction.

    It is represented by the following matrix for qubits:

    .. math::

        \\begin{pmatrix}
        \\frac{\\sqrt{2}}{2} & \\frac{\\sqrt{2}}{2} \\\\
        \\frac{\\sqrt{2}}{2} & -\\frac{\\sqrt{2}}{2} \\\\
        \\end{pmatrix}

    However, generally it is given by the following formula:

    .. math::
        H = \\frac{1}{\\sqrt(d)} \\sum_{ij} \\omega_d^{ij} |i\\rangle\\langle j|

    where

    .. math::
        \\omega = \\exp(\\frac{2\\pi i}{d})

    and `d` is the number of levels (2 levels is a qubit,
    3 levels is a qutrit, etc.)

    References:
        - https://www.frontiersin.org/articles/10.3389/fphy.2020.589504/full
        - https://pubs.aip.org/aip/jmp/article-abstract/56/3/032202/763827
    """

    _num_qudits = 1
    _qasm_name = 'h'

    def __init__(self, radix: int = 2) -> None:
        """
        Construct a HGate.

        Args:
            radix (int): The number of qudit levels (>=2). By default, this
                is 2, which is gives a qubit Hadamard gate.

        Raises:
            ValueError: if radix < 2
        """
        if not is_integer(radix):
            raise TypeError(f'Expected integer for radix, got {type(radix)}.')

        if radix < 2:
            raise ValueError(f'Radix must be greater than 1, got {radix}.')

        self._radix = radix

        # Calculate unitary
        if radix == 2:
            matrix = np.array(
                [
                    [np.sqrt(2) / 2, np.sqrt(2) / 2],
                    [np.sqrt(2) / 2, -np.sqrt(2) / 2],
                ],
                dtype=np.complex128,
            )
            self._utry = UnitaryMatrix(matrix)

        else:
            matrix = np.zeros([radix] * 2, dtype=np.complex128)
            omega = np.exp(2j * np.pi / radix)
            for i in range(radix):
                for j in range(i, radix):
                    val = omega ** (i * j)
                    matrix[i, j] = val
                    matrix[j, i] = val
            matrix *= 1 / np.sqrt(radix)
            self._utry = UnitaryMatrix(matrix, self.radixes)



"""
This subpackage implements Placement algorithms.

Placement algorithms are responsible for selecting which set of physical qubits
to start the circuit on.
"""
from __future__ import annotations

from bqskit.passes.mapping.placement.greedy import GreedyPlacementPass
from bqskit.passes.mapping.placement.trivial import TrivialPlacementPass

__all__ = ['GreedyPlacementPass', 'TrivialPlacementPass']



"""This module implements DifferentiableCostFunction base classes."""
from __future__ import annotations

import abc

import numpy as np
import numpy.typing as npt

from bqskit.ir.opt.cost.function import CostFunction
from bqskit.ir.opt.cost.residual import ResidualsFunction
from bqskit.qis.unitary.unitary import RealVector


class DifferentiableCostFunction(CostFunction):
    """
    The DifferentiableCostFunction base class.

    A DifferentiableCostFunction is a differentiable map from a vector of
    real numbers to a real number.

    A DifferentiableCostFunction exposes the `get_grad` abstract method
    and the `get_cost_and_grad` method. When subclassing, you only need to
    implement the `gen_cost` and `gen_grad` function factories. You can
    overwrite gen_cost_and_grad for optimization in some cases.
    """

    @abc.abstractmethod
    def get_grad(self, params: RealVector) -> npt.NDArray[np.float64]:
        """Return the cost gradient given the input parameters."""

    def get_cost_and_grad(
        self,
        params: RealVector,
    ) -> tuple[float, npt.NDArray[np.float64]]:
        """Return the cost and gradient given the input parameters."""
        return self.get_cost(params), self.get_grad(params)


class DifferentiableResidualsFunction(ResidualsFunction):
    """
    The DifferentiableResidualsFunction base class.

    A DifferentiableResidualsFunction is a differentiable map from a vector of
    real numbers to a matrix where the rows are the gradients for each
    input parameter.

    A DifferentiableResidualsFunction exposes the `get_grad` abstract method
    and the `get_residuals_and_grad` method. When subclassing, you only need to
    implement the `gen_cost` and `gen_grad` function factories. You can
    overwrite gen_cost_and_grad for optimization in some cases.
    """

    @abc.abstractmethod
    def get_grad(self, params: RealVector) -> npt.NDArray[np.float64]:
        """Return the residuals gradient given the input parameters."""

    def get_cost_and_grad(
        self,
        params: RealVector,
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        """Return the residuals and gradient given the input parameters."""
        return self.get_residuals(params), self.get_grad(params)



"""This module implements the MultiStartGenerator base class."""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    import numpy as np
    import numpy.typing as npt
    from bqskit.ir.circuit import Circuit
    from bqskit.qis.state.state import StateVector
    from bqskit.qis.state.system import StateSystem
    from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class MultiStartGenerator(abc.ABC):
    @abc.abstractmethod
    def gen_starting_points(
        self,
        multistarts: int,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector | StateSystem,
    ) -> list[npt.NDArray[np.float64]]:
        """
        Generate `multistarts` starting points for instantiation.

        Args:
            multistarts (int): The number of starting points to generate.

            circuit (Circuit): The circuit to generate the points for.

            target (UnitaryMatrix | StateVector | StateSystem): The target.

        Return:
            (list[npt.NDArray[np.float64]]): List of starting inputs for
                instantiation.

        Raises:
            ValueError: If `multistarts` is not a positive integer.
        """



"""
This package implements partitioning passes.

Partitioning passes group together adjacent gates into CircuitGates objects.
"""
from __future__ import annotations

from bqskit.passes.partitioning.cluster import ClusteringPartitioner
from bqskit.passes.partitioning.greedy import GreedyPartitioner
from bqskit.passes.partitioning.quick import QuickPartitioner
from bqskit.passes.partitioning.scan import ScanPartitioner
from bqskit.passes.partitioning.single import GroupSingleQuditGatePass

__all__ = [
    'ClusteringPartitioner',
    'GreedyPartitioner',
    'ScanPartitioner',
    'QuickPartitioner',
    'GroupSingleQuditGatePass',
]



"""This module implements the UnitaryBuilder class."""
from __future__ import annotations

import logging
from typing import cast
from typing import Sequence

import numpy as np
import numpy.typing as npt

from bqskit.ir.location import CircuitLocation
from bqskit.ir.location import CircuitLocationLike
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitary import Unitary
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_valid_radixes

logger = logging.getLogger(__name__)


class UnitaryBuilder(Unitary):
    """
    An object for fast unitary accumulation using tensor networks.

    A UnitaryBuilder is similar to a StringBuilder in the sense that it is an
    efficient way to string together or accumulate :class:`Unitary` objects.
    This class uses concepts from tensor networks to efficiently multiply
    unitary matrices.
    """

    def __init__(self, num_qudits: int, radixes: Sequence[int] = []) -> None:
        """
        UnitaryBuilder constructor.

        Args:
            num_qudits (int): The number of qudits to build a Unitary for.

            radixes (Sequence[int]): A sequence with its length equal
                to `num_qudits`. Each element specifies the base of a
                qudit. Defaults to qubits.

        Raises:
            ValueError: If `num_qudits` is nonpositive.

            ValueError: If the length of `radixes` is not equal to
                `num_qudits`.

        Examples:
            >>> builder = UnitaryBuilder(4)  # Creates a 4-qubit builder.
        """

        if not is_integer(num_qudits):
            raise TypeError(
                'Expected int for num_qudits, got %s.' %
                type(num_qudits),
            )

        if num_qudits <= 0:
            raise ValueError(
                'Expected positive number for num_qudits, got %d.' %
                num_qudits,
            )

        self._num_qudits = num_qudits
        self._radixes = tuple(radixes if len(radixes) > 0 else [2] * num_qudits)

        if not is_valid_radixes(self.radixes):
            raise TypeError('Invalid qudit radixes.')

        if len(self.radixes) != self.num_qudits:
            raise ValueError(
                'Expected length of radixes to be equal to num_qudits:'
                ' %d != %d' % (len(self.radixes), self.num_qudits),
            )

        self._num_params = 0
        self._dim = int(np.prod(self.radixes))
        self.tensor = np.identity(self.dim, dtype=np.complex128)
        self.tensor = self.tensor.reshape(self.radixes * 2)

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Build the unitary, see :func:`Unitary.get_unitary` for more."""
        utry = self.tensor.reshape((self.dim, self.dim))
        return UnitaryMatrix(utry, self.radixes, False)

    def apply_right(
        self,
        utry: UnitaryMatrix,
        location: CircuitLocationLike,
        inverse: bool = False,
        check_arguments: bool = True,
    ) -> None:
        """
        Apply the specified unitary on the right of this UnitaryBuilder.

        ..
                 .-----.   .------.
              0 -|     |---|      |-
              1 -|     |---| utry |-
                 .     .   '------'
                 .     .
                 .     .
            n-1 -|     |------------
                 '-----'

        Args:
            utry (UnitaryMatrix): The unitary to apply.

            location (CircuitLocationLike): The qudits to apply the unitary on.

            inverse (bool): If true, apply the inverse of the unitary.

            check_arguments (bool): If true, check the inputs for type and
                value errors.

        Raises:
            ValueError: If `utry`'s size does not match the given location.

            ValueError: if `utry`'s radixes does not match the given location.

        Notes:
            - Applying the unitary on the right is equivalent to multiplying
              the unitary on the left of the tensor. The notation comes
              from the quantum circuit perspective.

            - This operation is performed using tensor contraction.
        """

        if check_arguments:
            if not isinstance(utry, UnitaryMatrix):
                raise TypeError('Expected UnitaryMatrix, got %s', type(utry))

            if not CircuitLocation.is_location(location, self.num_qudits):
                raise TypeError('Invalid location.')

            location = CircuitLocation(location)

            if len(location) != utry.num_qudits:
                raise ValueError('Unitary and location size mismatch.')

            for utry_radix, bldr_radix_idx in zip(utry.radixes, location):
                if utry_radix != self.radixes[bldr_radix_idx]:
                    raise ValueError('Unitary and location radix mismatch.')

        left_perm = list(cast(CircuitLocation, location))
        mid_perm = [x for x in range(self.num_qudits) if x not in left_perm]
        right_perm = [x + self.num_qudits for x in range(self.num_qudits)]

        left_dim = int(np.prod([self.radixes[x] for x in left_perm]))

        utry = utry.dagger if inverse else utry

        perm = left_perm + mid_perm + right_perm
        self.tensor = self.tensor.transpose(perm)
        self.tensor = self.tensor.reshape((left_dim, -1))
        self.tensor = utry @ self.tensor

        shape = list(self.radixes) * 2
        shape = [shape[p] for p in perm]
        self.tensor = self.tensor.reshape(shape)
        inv_perm = list(np.argsort(perm))
        self.tensor = self.tensor.transpose(inv_perm)

    def apply_left(
        self,
        utry: UnitaryMatrix,
        location: CircuitLocationLike,
        inverse: bool = False,
        check_arguments: bool = True,
    ) -> None:
        """
        Apply the specified unitary on the left of this UnitaryBuilder.

        ..
                 .------.   .-----.
              0 -|      |---|     |-
              1 -| gate |---|     |-
                 '------'   .     .
                            .     .
                            .     .
            n-1 ------------|     |-
                            '-----'

        Args:
            utry (UnitaryMatrix): The unitary to apply.

            location (CircuitLocationLike): The qudits to apply the unitary on.

            inverse (bool): If true, apply the inverse of the unitary.

            check_arguments (bool): If true, check the inputs for type and
                value errors.

        Raises:
            ValueError: If `utry`'s size does not match the given location.

            ValueError: if `utry`'s radixes does not match the given location.

        Notes:
            - Applying the unitary on the left is equivalent to multiplying
              the unitary on the right of the tensor. The notation comes
              from the quantum circuit perspective.

            - This operation is performed using tensor contraction.
        """

        if check_arguments:
            if not isinstance(utry, UnitaryMatrix):
                raise TypeError('Expected UnitaryMatrix, got %s', type(utry))

            if not CircuitLocation.is_location(location, self.num_qudits):
                raise TypeError('Invalid location.')

            location = CircuitLocation(location)

            if len(location) != utry.num_qudits:
                raise ValueError('Unitary and location size mismatch.')

            for utry_radix, bldr_radix_idx in zip(utry.radixes, location):
                if utry_radix != self.radixes[bldr_radix_idx]:
                    raise ValueError('Unitary and location radix mismatch.')

        location = cast(CircuitLocation, location)
        left_perm = list(range(self.num_qudits))
        mid_perm = [
            x + self.num_qudits
            for x in left_perm
            if x not in location
        ]
        right_perm = [x + self.num_qudits for x in location]

        right_dim = int(
            np.prod([
                self.radixes[x - self.num_qudits]
                for x in right_perm
            ]),
        )

        utry = utry.dagger if inverse else utry

        perm = left_perm + mid_perm + right_perm
        self.tensor = self.tensor.transpose(perm)
        self.tensor = self.tensor.reshape((-1, right_dim))
        self.tensor = self.tensor @ utry

        shape = list(self.radixes) * 2
        shape = [shape[p] for p in perm]
        self.tensor = self.tensor.reshape(shape)
        inv_perm = list(np.argsort(perm))
        self.tensor = self.tensor.transpose(inv_perm)

    def eval_apply_right(
        self,
        M: npt.NDArray[np.complex128],
        location: CircuitLocationLike,
    ) -> npt.NDArray[np.complex128]:
        """
        Evaluate the application of `M` on the right of this UnitaryBuilder.

        See :func:`apply_right` for more info.
        """
        left_perm = list(cast(CircuitLocation, location))
        mid_perm = [x for x in range(self.num_qudits) if x not in left_perm]
        right_perm = [x + self.num_qudits for x in range(self.num_qudits)]

        left_dim = int(np.prod([self.radixes[x] for x in left_perm]))

        perm = left_perm + mid_perm + right_perm
        tensor_copy = self.tensor.copy()
        tensor_copy = tensor_copy.transpose(perm)
        tensor_copy = tensor_copy.reshape((left_dim, -1))
        tensor_copy = M @ tensor_copy  # TODO: Require out matrix to avoid copy

        shape = list(self.radixes) * 2
        shape = [shape[p] for p in perm]
        tensor_copy = tensor_copy.reshape(shape)
        inv_perm = list(np.argsort(perm))
        tensor_copy = tensor_copy.transpose(inv_perm)
        out_M = tensor_copy.reshape((self.dim, self.dim))
        return out_M

    def eval_apply_left(
        self,
        M: npt.NDArray[np.complex128],
        location: CircuitLocationLike,
    ) -> npt.NDArray[np.complex128]:
        """
        Evaluate the application of `M` on the left of this UnitaryBuilder.

        See :func:`apply_left` for more info.
        """
        location = cast(CircuitLocation, location)
        left_perm = list(range(self.num_qudits))
        mid_perm = [
            x + self.num_qudits
            for x in left_perm
            if x not in location
        ]
        right_perm = [x + self.num_qudits for x in location]

        right_dim = int(
            np.prod([
                self.radixes[x - self.num_qudits]
                for x in right_perm
            ]),
        )

        perm = left_perm + mid_perm + right_perm
        tensor_copy = self.tensor.copy()
        tensor_copy = tensor_copy.transpose(perm)
        tensor_copy = tensor_copy.reshape((-1, right_dim))
        tensor_copy = tensor_copy @ M

        shape = list(self.radixes) * 2
        shape = [shape[p] for p in perm]
        tensor_copy = tensor_copy.reshape(shape)
        inv_perm = list(np.argsort(perm))
        tensor_copy = tensor_copy.transpose(inv_perm)
        out_M = tensor_copy.reshape((self.dim, self.dim))
        return out_M

    def calc_env_matrix(
            self, location: Sequence[int],
    ) -> npt.NDArray[np.complex128]:
        """
        Calculates the environment matrix w.r.t. the specified location.

        Args:
            location (Sequence[int]): Calculate the environment matrix with
                respect to the qudit indices in location.

        Returns:
            np.ndarray: The environmental matrix.
        """

        left_perm = list(range(self.num_qudits))
        left_perm = [x for x in left_perm if x not in location]
        left_perm = left_perm + [x + self.num_qudits for x in left_perm]
        right_perm = list(location) + [x + self.num_qudits for x in location]

        perm = left_perm + right_perm
        a = np.transpose(self.tensor, perm)
        a = np.reshape(
            a, (
                2 ** (self.num_qudits - len(location)),
                2 ** (self.num_qudits - len(location)),
                2 ** len(location),
                2 ** len(location),
            ),
        )
        return np.trace(a)



"""This module implements helper methods for generating docs."""
from __future__ import annotations

import os


def building_docs() -> bool:
    """Return true if currently building documentations."""

    if 'BQSKIT_DOC_CHECK_OVERRIDE' in os.environ:
        return False
    if 'READTHEDOCS' in os.environ:
        return True
    if '__SPHINX_BUILD__' in os.environ:
        return True
    return False



"""This module implements the ApplyPlacement class."""
from __future__ import annotations

import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit

_logger = logging.getLogger(__name__)


class ApplyPlacement(BasePass):
    """Place the circuit on the machine model."""

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        model = data.model
        placement = data.placement

        # Place circuit on the model according to the placement
        physical_circuit = Circuit(model.num_qudits, model.radixes)
        physical_circuit.append_circuit(circuit, placement)
        circuit.become(physical_circuit)

        # Update the relevant data variables
        data.initial_mapping = [placement[p] for p in data.initial_mapping]
        data.final_mapping = [placement[p] for p in data.final_mapping]
        data.placement = list(i for i in range(model.num_qudits))



"""This module implements the DifferentiableUnitary base class."""
from __future__ import annotations

import abc

import numpy as np
import numpy.typing as npt

from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitary import Unitary
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class DifferentiableUnitary(Unitary):
    """
    A differentiable unitary-valued function.

    A `DifferentiableUnitary` inherits from `Unitary` and additionally
    exposes the :func:`get_grad` abstract method and the
    :func:`get_unitary_and_grad` method.
    """

    @abc.abstractmethod
    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for the unitary map as an np.ndarray.

        Args:
            params (RealVector): The unitary parameters, see
                :func:`Unitary.get_unitary` for more info.

        Returns:
            np.ndarray: The `(num_params,N,N)`-shaped, matrix-by-vector
            derivative of this unitary at the point specified by params.

        Notes:
            The gradient of a unitary is defined as a matrix-by-vector
            derivative. If the UnitaryMatrix result of `get_unitary` has
            dimension NxN, then the shape of `get_grad`'s return value
            should equal (num_params,N,N), where the return value's
            i-th element is the matrix derivative of the unitary
            with respect to the i-th parameter.
        """

    def get_unitary_and_grad(
        self,
        params: RealVector = [],
    ) -> tuple[UnitaryMatrix, npt.NDArray[np.complex128]]:
        """
        Return a tuple combining the outputs of `get_unitary` and `get_grad`.

        Args:
            params (RealVector): The unitary parameters, see
                :func:`Unitary.get_unitary` for more info.

        Returns:
            tuple: tuple containing:
                UnitaryMatrix: The unitary matrix, see
                :func:`Unitary.get_unitary` for more info.

                np.ndarray: The unitary's gradient, see :func:`get_grad`.

        Notes:
            Can be overridden to potentially speed up optimization by
            calculating both at the same time.
        """
        return (self.get_unitary(params), self.get_grad(params))



"""This module implements functions for translating to and from Qiskit."""
from __future__ import annotations

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from qiskit import QuantumCircuit

from qiskit import qasm2
from bqskit.ir.circuit import Circuit
from bqskit.ir.lang.qasm2 import OPENQASM2Language


def qiskit_to_bqskit(qc: QuantumCircuit) -> Circuit:
    """Convert Qiskit's QuantumCircuit `qc` to a BQSKit Circuit."""
    circuit = OPENQASM2Language().decode(qasm2.dumps(qc))
    # circuit.renumber_qudits(list(reversed(range(circuit.num_qudits))))
    return circuit
    # TODO: support gates not captured by qasm


def bqskit_to_qiskit(circuit: Circuit) -> QuantumCircuit:
    """Convert a BQSKit Circuit to Qiskit's QuantumCircuit."""
    try:
        from qiskit import QuantumCircuit
    except ImportError as e:
        raise ImportError(
            '\n\nUnable to import qiskit package.\n'
            'Please ensure that it is installed with the following command:\n'
            '\tpip install qiskit\n',
        ) from e

    # circuit.renumber_qudits(list(reversed(range(circuit.num_qudits))))
    return QuantumCircuit.from_qasm_str(OPENQASM2Language().encode(circuit))
    # TODO: support gates not captured by qasm



"""This module implements the BasePass abstract base class."""
from __future__ import annotations

import abc
import warnings
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any
    from bqskit.compiler.machine import MachineModel
    from bqskit.compiler.passdata import PassData
    from bqskit.compiler.workflow import Workflow
    from bqskit.ir.circuit import Circuit
    from bqskit.qis.graph import CouplingGraph
    from bqskit.qis.state.system import StateSystem
    from bqskit.qis.state.state import StateVector
    from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class BasePass(abc.ABC):
    """
    The abstract base for BQSKit compiler passes.

    All BQSKit algorithms must inherit from BasePass to be run within
    the compiler framework. Each child class will need to implement its
    algorithm inside of the :func:`run` method.

    Examples:
        >>> class PrintCNOTCountPass(BasePass):
        ...     async def run(self, circuit: Circuit, data: PassData) -> None:
        ...         print(f"Number of CNOTs: {circuit.count(CNOTGate())}")
    """

    @property
    def name(self) -> str:
        """The name of the pass."""
        return self.__class__.__name__

    @abc.abstractmethod
    async def run(self, circuit: Circuit, data: PassData) -> None:
        """
        Perform the pass's operation on `circuit`.

        Args:
            circuit (Circuit): The circuit to operate on.

            data (PassData): Associated data for the pass.
                Can be used to get auxillary information from previous
                passes and to store information for future passes.

        Note:
            - This function should be self-contained and have no side effects.
              This is because it will be potentially run in parallel.
        """

    def __str__(self) -> str:
        """Return a string representation of the pass."""
        return self.name

    @staticmethod
    def get_model(_: Any, data: PassData) -> MachineModel:
        """
        Retrieve the machine model from the data dictionary.

        (Deprecated)
        """
        warnings.warn(
            'BasePass calls to retrieve elements from the data dictionary'
            ' are now deprecated. We have upgraded the features of the data'
            ' dictionary. It can now be used directly to retrieve elements'
            " and provide defaults if they don't exist. This warning will"
            ' become an error in the future.',
            DeprecationWarning,
        )
        return data.model

    @staticmethod
    def get_placement(_: Any, data: PassData) -> list[int]:
        """
        Retrieve the logical to physical qubit map from the data dictionary.

        (Deprecated)
        """
        warnings.warn(
            'BasePass calls to retrieve elements from the data dictionary'
            ' are now deprecated. We have upgraded the features of the data'
            ' dictionary. It can now be used directly to retrieve elements'
            " and provide defaults if they don't exist. This warning will"
            ' become an error in the future.',
            DeprecationWarning,
        )
        return data.placement

    @staticmethod
    def get_connectivity(_: Any, data: PassData) -> CouplingGraph:
        """
        Retrieve the current connectivity of the circuit.

        (Deprecated)
        """
        warnings.warn(
            'BasePass calls to retrieve elements from the data dictionary'
            ' are now deprecated. We have upgraded the features of the data'
            ' dictionary. It can now be used directly to retrieve elements'
            " and provide defaults if they don't exist. This warning will"
            ' become an error in the future.',
            DeprecationWarning,
        )
        return data.connectivity

    @staticmethod
    def get_target(
        _: Any,
        data: PassData,
    ) -> UnitaryMatrix | StateVector | StateSystem:
        """
        Retrieve the target from the data dictionary.

        (Deprecated)
        """
        warnings.warn(
            'BasePass calls to retrieve elements from the data dictionary'
            ' are now deprecated. We have upgraded the features of the data'
            ' dictionary. It can now be used directly to retrieve elements'
            " and provide defaults if they don't exist. This warning will"
            ' become an error in the future.',
            DeprecationWarning,
        )
        return data.target

    @staticmethod
    def in_parallel(data: dict[str, Any]) -> bool:
        """
        Return true if pass is being executed in a parallel.

        (Deprecated)
        """
        warnings.warn(
            'BasePass calls to `in_parallel` are deprecated and will always'
            ' return True. This warning will become an error in the future.',
            DeprecationWarning,
        )
        return True

    @staticmethod
    def execute(*args: Any, **kwargs: Any) -> Any:
        """
        Map a function over iterable arguments in parallel.

        (Defunct)
        """
        raise RuntimeError(
            'Since Dask was removed, the execute function became defunct.'
            ' We have switched to using `get_runtime().map(...)`. You can'
            ' import `get_runtime` from `bqskit.runtime`, and alongside a'
            ' an `await` keyword, should be a  drop-in replacement for the'
            ' execute function. See the following link for more info: '
            'https://bqskit.readthedocs.io/en/latest/source/runtime.html'
            'In a future version, this error will become an AttributeError.',
        )


async def _sub_do_work(
    workflow: Workflow,
    circuit: Circuit,
    data: PassData,
) -> tuple[Circuit, PassData]:
    """Execute a sequence of passes on circuit."""
    if 'calculate_error_bound' in data and data['calculate_error_bound']:
        old_utry = circuit.get_unitary()

    await workflow.run(circuit, data)

    if 'calculate_error_bound' in data and data['calculate_error_bound']:
        new_utry = circuit.get_unitary()
        data.error = new_utry.get_distance_from(old_utry)

    return circuit, data



"""This module implements the DetachedServer runtime."""
from __future__ import annotations

import argparse
import logging
import selectors
import socket
import time
import uuid
from dataclasses import dataclass
from logging import LogRecord
from multiprocessing.connection import Connection
from multiprocessing.connection import Listener
from threading import Thread
from typing import Any
from typing import cast
from typing import List
from typing import Sequence

from bqskit.compiler.status import CompilationStatus
from bqskit.compiler.task import CompilationTask
from bqskit.runtime import default_server_port
from bqskit.runtime.address import RuntimeAddress
from bqskit.runtime.base import import_tests_package
from bqskit.runtime.base import parse_ipports
from bqskit.runtime.base import ServerBase
from bqskit.runtime.direction import MessageDirection
from bqskit.runtime.message import RuntimeMessage
from bqskit.runtime.result import RuntimeResult
from bqskit.runtime.task import RuntimeTask


def listen(server: DetachedServer, port: int) -> None:
    """Listening thread listens for client connections."""
    listener = Listener(('0.0.0.0', port))
    while server.running:
        client = listener.accept()

        if server.running:
            # We check again that the server is running before registering
            # the client because dummy data is sent to unblock
            # listener.accept() during server shutdown
            server.clients[client] = set()
            server.sel.register(
                client,
                selectors.EVENT_READ,
                MessageDirection.CLIENT,
            )
            server.logger.debug('Connected and registered new client.')

    listener.close()


@dataclass
class ServerMailbox:
    """A mailbox on a server is a final destination for a compilation task."""

    result: Any = None
    """Where the mailbox will store the result once it arrives."""

    client_waiting: bool = False
    """If true, the server knows a client is blocked waiting on this result."""

    @property
    def ready(self) -> bool:
        """Return true if the mailbox result is ready and waiting delivery."""
        return self.result is not None


class DetachedServer(ServerBase):
    """
    BQSKit Runtime Server in detached mode.

    In detached mode, the runtime is started separately from the client. Clients
    can connect and disconnect while not shutting down the server. This
    architecture is designed for the distributed setting, where managers manage
    workers in shared memory and communicate with the server over a network.
    """

    def __init__(
        self,
        ipports: Sequence[tuple[str, int]],
        port: int = default_server_port,
    ) -> None:
        """
        Create a server and connect to the managers at `ipports`.

        Args:
            ipports (list[tuple[str, int]]): The ip and port pairs were
                managers are expected to be listening for server connections.

            port (int): The port this server will listen for clients on.
                Default can be found in the
                :obj:`~bqskit.runtime.default_server_port` global variable.
        """
        super().__init__()

        self.clients: dict[Connection, set[uuid.UUID]] = {}
        """Tracks all connected clients and all the tasks they have created."""

        self.tasks: dict[uuid.UUID, tuple[int, Connection]] = {}
        """Tracks all active CompilationTasks submitted to the cluster."""

        self.mailbox_to_task_dict: dict[int, uuid.UUID] = {}
        """Used to convert internal RuntimeTasks to client CompilationTasks."""

        self.mailboxes: dict[int, ServerMailbox] = {}
        """Mapping from mailbox ids to mailboxes."""

        self.mailbox_counter = 0
        """Counter to ensure all mailboxes have unique IDs."""

        # Connect to managers
        self.connect_to_managers(ipports)

        # Start client listener
        self.port = port
        self.listen_thread = Thread(target=listen, args=(self, port))
        self.listen_thread.start()
        self.logger.info(f'Started client listener on port {self.port}.')

    def handle_message(
        self,
        msg: RuntimeMessage,
        direction: MessageDirection,
        conn: Connection,
        payload: Any,
    ) -> None:
        """Process the message coming from `direction`."""
        if direction == MessageDirection.CLIENT:

            if msg == RuntimeMessage.CONNECT:
                pass

            elif msg == RuntimeMessage.DISCONNECT:
                self.handle_disconnect(conn)

            elif msg == RuntimeMessage.SUBMIT:
                ctask = cast(CompilationTask, payload)
                self.handle_new_comp_task(conn, ctask)

            elif msg == RuntimeMessage.REQUEST:
                request = cast(uuid.UUID, payload)
                self.handle_request(conn, request)

            elif msg == RuntimeMessage.STATUS:
                request = cast(uuid.UUID, payload)
                self.handle_status(conn, request)

            elif msg == RuntimeMessage.CANCEL:
                request = cast(uuid.UUID, payload)
                self.handle_cancel_comp_task(request)

            else:
                raise RuntimeError(f'Unexpected message type: {msg.name}')

        elif direction == MessageDirection.BELOW:

            if msg == RuntimeMessage.SUBMIT:
                rtask = cast(RuntimeTask, payload)
                self.schedule_tasks([rtask])

            elif msg == RuntimeMessage.SUBMIT_BATCH:
                rtasks = cast(List[RuntimeTask], payload)
                self.schedule_tasks(rtasks)

            elif msg == RuntimeMessage.RESULT:
                result = cast(RuntimeResult, payload)
                self.handle_result(result)

            elif msg == RuntimeMessage.ERROR:
                self.handle_error(payload)
                return

            elif msg == RuntimeMessage.LOG:
                self.handle_log(payload)

            elif msg == RuntimeMessage.CANCEL:
                self.broadcast_cancel(payload)

            elif msg == RuntimeMessage.SHUTDOWN:
                self.handle_shutdown()

            elif msg == RuntimeMessage.WAITING:
                num_idle = cast(int, payload)
                self.handle_waiting(conn, num_idle)

            elif msg == RuntimeMessage.UPDATE:
                task_diff = cast(int, payload)
                self.conn_to_employee_dict[conn].num_tasks += task_diff

            else:
                raise RuntimeError(f'Unexpected message type: {msg.name}')

        else:
            raise RuntimeError(f'Unexpected message from {direction.name}.')

    def handle_system_error(self, error_str: str) -> None:
        """
        Handle an error in runtime code as opposed to client code.

        This is called when an error arises in runtime code not in a
        RuntimeTask's coroutine code.
        """
        for client in self.clients.keys():
            client.send((RuntimeMessage.ERROR, error_str))

        # Sleep to ensure clients receive error message before shutdown
        time.sleep(1)

    def handle_shutdown(self) -> None:
        """Shutdown the runtime."""
        super().handle_shutdown()

        # Close client connections
        for client in self.clients.keys():
            try:
                client.close()
            except Exception:
                pass
        self.clients.clear()
        self.logger.debug('Cleared clients.')

        # Close listener (hasattr checked for attachedserver shutdown)
        if hasattr(self, 'listen_thread') and self.listen_thread.is_alive():
            # Listener will be blocked on the accept call
            # Create a dummy connection to unblock it
            # Workaround credit: https://stackoverflow.com/questions/16734534
            # Related bug: https://github.com/python/cpython/issues/76425
            dummy_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            dummy_socket.connect(('localhost', self.port))
            dummy_socket.close()
            self.listen_thread.join()
            self.logger.debug('Joined listening thread.')

    def handle_disconnect(self, conn: Connection) -> None:
        """Disconnect a client connection from the runtime."""
        super().handle_disconnect(conn)
        tasks = self.clients.pop(conn)
        for task_id in tasks:
            self.handle_cancel_comp_task(task_id)
        self.logger.info('Unregistered client.')

    def handle_new_comp_task(
        self,
        conn: Connection,
        task: CompilationTask,
    ) -> None:
        """Convert a :class:`CompilationTask` into an internal one."""
        mailbox_id = self._get_new_mailbox_id()
        self.tasks[task.task_id] = (mailbox_id, conn)
        self.mailbox_to_task_dict[mailbox_id] = task.task_id
        self.mailboxes[mailbox_id] = ServerMailbox()
        self.logger.info(f'New CompilationTask: {task.task_id}.')

        self.clients[conn].add(task.task_id)

        internal_task = RuntimeTask(
            (CompilationTask.run, (task,), {}),
            RuntimeAddress(-1, mailbox_id, 0),
            mailbox_id,
            tuple(),
            task.logging_level,
            task.max_logging_depth,
        )

        self.schedule_tasks([internal_task])

    def handle_request(self, conn: Connection, request: uuid.UUID) -> None:
        """Record the requested task, and ship it as soon as it's ready."""
        if request not in self.clients[conn] or request not in self.tasks:
            self.outgoing.put((conn, RuntimeMessage.ERROR, 'Unknown task.'))
            self.handle_disconnect(conn)  # Bad client
            return

        # Get the mailbox associated with this task.
        mailbox_id = self.tasks[request][0]
        box = self.mailboxes[mailbox_id]

        if box.ready:
            # If the result has already arrived, ship it to the client.
            self.logger.info(f'Responding to request for task {request}.')
            self.outgoing.put((conn, RuntimeMessage.RESULT, box.result))
            self.mailboxes.pop(mailbox_id)
            self.clients[conn].remove(request)
            # t_id is not removed from self.tasks or
            # self.mailbox_to_task_dict incase there are left
            # over log messages arriving.

        else:
            # Otherwise, note that the client is waiting on this result.
            # This is so the result will be shipped the instant it arrives.
            box.client_waiting = True

    def handle_status(self, conn: Connection, request: uuid.UUID) -> None:
        """Inform the client if the task is finished or not."""
        if request not in self.clients[conn] or request not in self.tasks:
            # This task is unknown to the system
            m = (conn, RuntimeMessage.STATUS, CompilationStatus.UNKNOWN)
            self.outgoing.put(m)

        # Get the mailbox associated with this task.
        mailbox_id = self.tasks[request][0]
        box = self.mailboxes[mailbox_id]

        # Send status
        s = CompilationStatus.DONE if box.ready else CompilationStatus.RUNNING
        self.outgoing.put((conn, RuntimeMessage.STATUS, s))

    def handle_cancel_comp_task(self, request: uuid.UUID) -> None:
        """Cancel a compilation task in the system."""
        self.logger.info(f'Cancelling: {request}.')

        # Remove task from server data
        mailbox_id, client_conn = self.tasks[request]
        self.mailboxes.pop(mailbox_id)
        if client_conn in self.clients:
            self.clients[client_conn].remove(request)

        # Forward internal cancel messages
        addr = RuntimeAddress(-1, mailbox_id, 0)
        self.broadcast_cancel(addr)

        # Acknowledge the client's cancel request
        if not client_conn.closed:
            # Check if it closed first since the client may have disconnected
            self.outgoing.put((client_conn, RuntimeMessage.CANCEL, None))

    def handle_result(self, result: RuntimeResult) -> None:
        """Either store the result here or ship it to the destination worker."""
        # Record a task has been completed
        self.get_employee_responsible_for(result.completed_by).num_tasks -= 1

        # Check if the result is for a client
        if result.return_address.worker_id == -1:
            mailbox_id = result.return_address.mailbox_index
            if mailbox_id not in self.mailboxes:
                return  # Silently discard results from cancelled tasks

            box = self.mailboxes[mailbox_id]
            box.result = result.result
            t_id = self.mailbox_to_task_dict[mailbox_id]
            self.logger.info(f'Finished: {t_id}.')

            if box.client_waiting:
                self.logger.info(f'Responding to request for task {t_id}.')
                m = (self.tasks[t_id][1], RuntimeMessage.RESULT, box.result)
                self.outgoing.put(m)
                self.clients[self.tasks[t_id][1]].remove(t_id)
                self.mailboxes.pop(mailbox_id)
                # t_id is not removed from self.tasks or
                # self.mailbox_to_task_dict incase there are left
                # over log messages arriving.

        else:
            self.send_result_down(result)

    def handle_error(self, error_payload: tuple[int, str]) -> None:
        """Forward an error to the appropriate client and disconnect it."""
        if not isinstance(error_payload, tuple):
            # Internal errors may bubble up without a task_id
            assert isinstance(error_payload, str)
            self.handle_system_error(error_payload)
            self.handle_shutdown()
            raise RuntimeError(error_payload)

        tid = error_payload[0]
        conn = self.tasks[self.mailbox_to_task_dict[tid]][1]
        self.outgoing.put((conn, RuntimeMessage.ERROR, error_payload[1]))

    def handle_log(self, log_payload: tuple[int, LogRecord]) -> None:
        """Forward logs to appropriate client."""
        tid = log_payload[0]
        conn = self.tasks[self.mailbox_to_task_dict[tid]][1]
        self.outgoing.put((conn, RuntimeMessage.LOG, log_payload[1]))

    def _get_new_mailbox_id(self) -> int:
        """Unique mailbox id counter."""
        new_id = self.mailbox_counter
        self.mailbox_counter += 1
        return new_id


def start_server() -> None:
    """Entry point for a detached runtime server process."""
    parser = argparse.ArgumentParser(
        prog='bqskit-server',
        description='Launch a BQSKit runtime server process.',
    )
    parser.add_argument(
        'managers',
        nargs='+',
        help='The ip and port pairs were managers are expected to be waiting.',
    )
    parser.add_argument(
        '-p', '--port',
        type=int,
        default=default_server_port,
        help='The port this server will listen for clients on.',
    )
    parser.add_argument(
        '--verbose', '-v',
        action='count',
        default=0,
        help='Enable logging of increasing verbosity, either -v, -vv, or -vvv.',
    )
    parser.add_argument(
        '--import-tests', '-i',
        action='store_true',
        help='Import the bqskit tests package; used during testing.',
    )
    args = parser.parse_args()
    ipports = parse_ipports(args.managers)

    # Set up logging
    _logger = logging.getLogger('bqskit-runtime')
    _logger.setLevel([30, 20, 10, 1][min(args.verbose, 3)])
    _logger.addHandler(logging.StreamHandler())

    # Import tests package recursively
    if args.import_tests:
        import_tests_package()

    # Create the server
    server = DetachedServer(ipports, args.port)

    # Start the server
    server.run()



"""This module implements the QuestRunner CircuitRunner."""
from __future__ import annotations

import logging
import time
from typing import Any

import numpy as np
import numpy.typing as npt
from scipy.optimize import dual_annealing

from bqskit.compiler.compiler import Compiler
from bqskit.compiler.machine import MachineModel
from bqskit.exec.results import RunnerResults
from bqskit.exec.runner import CircuitRunner
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.circuitgate import CircuitGate
from bqskit.ir.gates.constant.cx import CNOTGate
from bqskit.ir.operation import Operation
from bqskit.ir.point import CircuitPoint
from bqskit.passes.control import ForEachBlockPass
from bqskit.passes.mapping.setmodel import SetModelPass
from bqskit.passes.partitioning import QuickPartitioner
from bqskit.passes.synthesis import LEAPSynthesisPass

_logger = logging.getLogger(__name__)


class QuestRunner(CircuitRunner):
    """
    Simulate a circuit using the Quest algorithm.

    References:
        Patel, Tirthak, et al. "Robust and Resource-Efficient Quantum Circuit
        Approximation." arXiv preprint arXiv:2108.12714 (2021).
    """

    def __init__(
        self,
        sub_runner: CircuitRunner,
        block_size: int = 3,
        compiler: Compiler | None = None,
        weit: float = 0.5,
        pert: int = 100,
        approx_threshold: float | None = None,
        sample_size: int = 16,
        model: MachineModel | None = None,
        instantiate_options: dict[str, Any] = {},
    ) -> None:
        """
        Run the QUEST algorithm using `sub_runner` to execute circuits.

        Args:
            sub_runner (CircuitRunner): The CircuitRunner to execute the
                approximated circuits.

            block_size (int): The maximum number of qudits in each
                block. Defaults to 3.

            compiler (Compiler | None): The compiler object used
                to partitioning and synthesize the circuit. If left
                as None, then a standard compiler will be created.

            weit (float): The weight set on number of CNOTs versus
                approximation dissimilarity. Default 0.5.

            pert (int): The percentile of sample distances used to judge
                approximation quality.

            approx_threshold (float | None): The maximum allowed error
                in the approximation. If left as None, then it will be
                computed based on the number of blocks.

            sample_size (int): The number of approximations to generate.

            model (MachineModel | None): The model to compile to.

            instantiate_options (dict[str, Any]): Options passed directly
                to instantiation.
        """
        self.sub_runner = sub_runner
        self.block_size = block_size
        self.compiler = compiler
        self.weit = weit
        self.pert = pert
        self.approx_threshold = approx_threshold
        self.sample_size = sample_size
        self.model = model
        self.instantiate_options = instantiate_options

    def run(self, circuit: Circuit) -> RunnerResults:
        """Execute the circuit, see CircuitRunner.run for more info."""

        # 1. Compile the circuit
        synthesis_pass = LEAPSynthesisPass(
            store_partial_solutions=True,
            instantiate_options=self.instantiate_options,
        )
        pass_list = [
            QuickPartitioner(self.block_size),
            ForEachBlockPass(synthesis_pass),
        ]
        if self.model is not None:
            pass_list.insert(0, SetModelPass(self.model))

        started = False
        if self.compiler is None:
            self.compiler = Compiler()
            started = True

        blocked_circuit, data = self.compiler.compile(
            circuit.copy(), pass_list, True,
        )

        # 2. Gather partial solutions
        data = data[ForEachBlockPass.key]
        psols, pts = self.parse_data(blocked_circuit, data)
        # psols: psols[i] = list[[circuit, dist]] -> block i's partial solutions
        # pts: pts[i] = CircuitPoint -> block i's locations

        if started:
            self.compiler.close()
            self.compiler = None

        # 3. Approximate circuit
        approx_circuits = self.approximate_circuit(blocked_circuit, psols, pts)

        # 4. Run circuits
        results = [self.sub_runner.run(c) for c in approx_circuits]

        # 5. Average and return results
        probs = np.sum(np.array([result.probs for result in results]), axis=0)
        probs /= self.sample_size
        return RunnerResults(circuit.num_qudits, circuit.radixes, probs)

    def parse_data(
        self,
        blocked_circuit: Circuit,
        data: dict[Any, Any],
    ) -> tuple[list[list[tuple[Circuit, float]]], list[CircuitPoint]]:
        """Parse the data outputed from synthesis."""
        block_data = data[0]

        psols: list[list[tuple[Circuit, float]]] = []
        pts: list[CircuitPoint] = []

        for block in block_data:
            pts.append(block['point'])
            exact_block = blocked_circuit[pts[-1]].gate._circuit.copy()  # type: ignore  # noqa
            exact_block.set_params(blocked_circuit[pts[-1]].params)
            exact_utry = exact_block.get_unitary()
            psols.append([(exact_block, 0.0)])

            if 'psols' not in block:
                continue

            for depth, psol_list in block['psols'].items():
                for psol in psol_list:
                    dist = psol[0].get_unitary().get_distance_from(exact_utry)
                    psols[-1].append((psol[0], dist))

        return psols, pts

    def approximate_circuit(
        self,
        circuit: Circuit,
        psols: list[list[tuple[Circuit, float]]],
        pts: list[CircuitPoint],
    ) -> list[Circuit]:
        """Use partial block solutions to compute circuit approximations."""
        num_blocks = len(pts)
        approx_threshold = self.approx_threshold or num_blocks / 10
        bounds = [[0, len(psol_list)] for psol_list in psols]

        psol_configs: list[list[int]] = []
        approx_circuits: list[Circuit] = []
        dists = [[psol[1] for psol in psol_list] for psol_list in psols]
        utrys = [
            [psol[0].get_unitary() for psol in psol_list]
            for psol_list in psols
        ]
        n_cxs = [
            [psol[0].count(CNOTGate()) for psol in psol_list]
            for psol_list in psols
        ]
        n_cx = sum(op.gate._circuit.count(CNOTGate()) for op in circuit)  # type: ignore  # noqa

        for _ in range(self.sample_size):
            x = []
            f = float('inf')
            for _ in range(3):
                result = dual_annealing(
                    annealing_objective,
                    args=(
                        approx_threshold,
                        n_cx,
                        psol_configs,
                        dists,
                        utrys,
                        n_cxs,
                        self.pert,
                        self.weit,
                    ),
                    x0=[0] * num_blocks,
                    bounds=bounds,
                    seed=int(time.time()),
                    initial_temp=5e4,
                    maxiter=10000,
                )

                if result.fun < f:
                    x = result.x
                    f = result.fun

            blocks = [int(x_e) for x_e in x]

            if blocks in psol_configs:
                _logger.debug('Generated a repeat approximate circuit.')
                break

            approx_dist = sum(
                psols[b][ind][1]
                for b, ind in enumerate(blocks)
            )

            if approx_dist > approx_threshold:
                _logger.debug('Approximate circuit has high distance.')
                break

            psol_configs.append(blocks)
            approx_circuit = self.assemble_circuit(circuit, blocks, psols, pts)
            approx_circuit.unfold_all()
            approx_circuits.append(approx_circuit)
            _logger.info(
                'Generated approximate circuit with approximate distance:'
                f' {approx_dist}.',
            )
            _logger.info(
                f'Approximate circuit has {approx_circuit.count(CNOTGate())}'
                ' cnots.',
            )

        return approx_circuits

    def assemble_circuit(
        self,
        circuit: Circuit,
        blocks: list[int],
        psols: list[list[tuple[Circuit, float]]],
        pts: list[CircuitPoint],
    ) -> Circuit:
        """Assemble a circuit from a list of block indices."""
        circuit_gates = [
            CircuitGate(psol_list[b][0])
            for b, psol_list
            in zip(blocks, psols)
        ]
        locations = [circuit[pt].location for pt in pts]
        operations = [
            Operation(cg, loc, cg._circuit.params)
            for cg, loc
            in zip(circuit_gates, locations)
        ]
        copied_circuit = circuit.copy()
        copied_circuit.batch_replace(pts, operations)
        return copied_circuit


def annealing_objective(x: npt.NDArray[np.float64], *args: Any) -> float:
    """Compute the objective function used during approximation generation."""
    blocks = [int(ind) for ind in x]

    approx_threshold = args[0]
    upper_limit = args[1]
    psol_configs = args[2]
    dists = args[3]
    utrys = args[4]
    n_cxs = args[5]
    p = args[6]
    w = args[7]

    if blocks in psol_configs:
        return 1.2

    approx_dist = sum(dists[i][b] for i, b in enumerate(blocks))

    if approx_dist > approx_threshold:
        return 1.1

    n_cx = w * (sum(n_cxs[i][b] for i, b in enumerate(blocks)) / upper_limit)

    if len(psol_configs) == 0:
        return n_cx

    distances = [0] * len(psol_configs)

    for index, config in enumerate(psol_configs):
        distances[index] = sum(
            utrys[i][config[i]].get_distance_from(utrys[i][b])
            < max(dists[i][config[i]], dists[i][b])
            for i, b in enumerate(blocks)
        ) / len(blocks)

    b_dv = (1 - w) * np.percentile(distances, p)

    return n_cx + b_dv


def gen_approximate_circuits(
    circuit: Circuit,
    block_size: int = 3,
    compiler: Compiler | None = None,
    weit: float = 0.5,
    pert: int = 100,
    approx_threshold: float | None = None,
    sample_size: int = 16,
    model: MachineModel | None = None,
    instantiate_options: dict[str, Any] = {},
) -> list[Circuit]:
    """
    Use the QUEST algorithm to generate approximate circuits.

    Args:
        block_size (int): The maximum number of qudits in each
            block. Defaults to 3.

        compiler (Compiler | None): The compiler object used
            to partitioning and synthesize the circuit. If left
            as None, then a standard compiler will be created.

        weit (float): The weight set on number of CNOTs versus
            approximation dissimilarity. Default 0.5.

        pert (int): The percentile of sample distances used to judge
            approximation quality.

        approx_threshold (float | None): The maximum allowed error
            in the approximation. If left as None, then it will be
            computed based on the number of blocks.

        sample_size (int): The number of approximations to generate.

        model (MachineModel | None): The model to compile to.

        instantiate_options (dict[str, Any]): Options passed directly
            to instantiation.
    """
    # 1. Compile the circuit
    synthesis_pass = LEAPSynthesisPass(
        store_partial_solutions=True,
        instantiate_options=instantiate_options,
    )
    pass_list = [
        QuickPartitioner(block_size),
        ForEachBlockPass(synthesis_pass),
    ]
    if model is not None:
        pass_list.insert(0, SetModelPass(model))

    started = False
    if compiler is None:
        compiler = Compiler()
        started = True

    blocked_circuit, data = compiler.compile(circuit.copy(), pass_list, True)
    data = data[ForEachBlockPass.key]

    # 2. Gather partial solutions
    runner = QuestRunner(
        None,  # type: ignore
        block_size,
        compiler,
        weit,
        pert,
        approx_threshold,
        sample_size,
        model,
        instantiate_options,
    )

    psols, pts = runner.parse_data(blocked_circuit, data)
    # psols: psols[i] = list[[circuit, dist]] -> block i's partial solutions
    # pts: pts[i] = CircuitPoint -> block i's locations

    # 3. Approximate circuit
    approx_circuits = runner.approximate_circuit(blocked_circuit, psols, pts)

    if started:
        compiler.close()

    return approx_circuits



"""This module implements the SwapGate."""
from __future__ import annotations

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.qis.permutation import PermutationMatrix
from bqskit.utils.typing import is_integer


class SwapGate(ConstantGate):
    """
    The two-qudit swap gate.

    This gate swaps the state of two qudits. For example, The qubit swap
    gate is given by the following unitary:

    .. math::

        \\begin{pmatrix}
        1 & 0 & 0 & 0 \\\\
        0 & 0 & 1 & 0 \\\\
        0 & 1 & 0 & 0 \\\\
        0 & 0 & 0 & 1 \\\\
        \\end{pmatrix}
    """

    def __init__(self, radix: int = 2) -> None:
        """
        Create a swap gate, defaulting to the qubit swap gate.

        Args:
            radix (int): The base of the qudits being swapped.
                Defaults to qubits or base 2. (Default: 2)

        Raises:
            ValueError: If radix is less than two.
        """
        if not is_integer(radix):
            raise TypeError('Expected a single integer radix.')

        if radix < 2:
            raise ValueError('Radix must be at least 2.')

        self._num_qudits = 2
        self._radixes = (radix, radix)
        self._dim = radix * radix
        self._utry = PermutationMatrix.gen_swap_unitary(radix)
        self._qasm_name = 'swap'

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, SwapGate)
            and self.radixes == other.radixes
        )

    def __hash__(self) -> int:
        return hash(('swapgate', self.radixes[0]))

    def __str__(self) -> str:
        if self.is_qubit_only():
            return 'SwapGate'
        else:
            return f'SwapGate({self.radixes[0]})'



"""This module implements MachineModels for Rigetti's QPUs."""
from __future__ import annotations

from bqskit.compiler.machine import MachineModel
from bqskit.ir.gates.constant.cz import CZGate
from bqskit.ir.gates.constant.sx import SXGate
from bqskit.ir.gates.constant.x import XGate
from bqskit.ir.gates.parameterized.rz import RZGate
from bqskit.qis.graph import CouplingGraph

rigetti_gate_set = {SXGate(), XGate(), RZGate(), CZGate()}

_aspen_11_coupling_graph = CouplingGraph([
    # Ring 1
    (0, 1), (1, 2), (2, 3), (3, 4),
    (4, 5), (5, 6), (6, 7), (7, 0),

    # Ring 2
    (8, 9), (9, 10), (10, 11), (11, 12),
    (12, 13), (13, 14), (14, 15), (15, 8),

    # Ring 3
    (16, 17), (17, 18), (19, 20),
    (20, 21), (21, 22), (22, 23), (23, 16),

    # Ring 4
    (24, 25), (25, 26), (26, 27), (27, 28),
    (28, 29), (29, 30), (30, 31), (31, 24),

    # Ring 5
    (34, 35), (35, 36),
    (36, 37), (37, 38), (38, 39),

    # Ring 1-2
    (1, 14), (2, 13),

    # Ring 2-3
    (9, 22), (10, 21),

    # Ring 3-4
    (17, 30),

    # Ring 4-5
    (25, 38), (26, 37),
])
"""Retrieved August 31, 2022: https://qcs.rigetti.com/qpus."""

_octo_rings = [
    [
        (i + 0, i + 1), (i + 1, i + 2), (i + 2, i + 3), (i + 3, i + 4),
        (i + 4, i + 5), (i + 5, i + 6), (i + 6, i + 7), (i + 7, i + 0),
    ]
    for i in range(0, 10 * 8, 8)
]

_horizontal_connections = [
    [
        (i + 1, j + 6),
        (i + 2, j + 5),
    ]
    for i, j in [(8 * r, 8 * (r + 1)) for r in [0, 1, 2, 3, 5, 6, 7, 8]]
]

_vertical_connections = [
    [
        (i + 0, j + 3),
        (i + 7, j + 4),
    ]
    for i, j in [(8 * r, 8 * (r + 5)) for r in range(5)]
]
_links = []
for l in _octo_rings:
    _links.extend(l)
for l in _horizontal_connections:
    _links.extend(l)
for l in _vertical_connections:
    _links.extend(l)

_aspen_m2_coupling_graph = CouplingGraph(_links)
"""Retrieved August 31, 2022: https://qcs.rigetti.com/qpus."""

Aspen11Model = MachineModel(40, _aspen_11_coupling_graph, rigetti_gate_set)
"""A BQSKit MachineModel for Rigetti's Aspen-11 quantum processor."""

AspenM2Model = MachineModel(80, _aspen_m2_coupling_graph, rigetti_gate_set)
"""A BQSKit MachineModel for Rigetti's Aspen-M-2 quantum processor."""

__all__ = ['Aspen11Model', 'AspenM2Model']



"""This module contains the version information for BQSKit."""
from __future__ import annotations
__version_info__ = ('1', '1', '2')
__version__ = '.'.join(__version_info__[:3]) + ''.join(__version_info__[3:])



"""This module implements the RuntimeFuture class."""
from __future__ import annotations

from typing import Any


class RuntimeFuture:
    """
    An awaitable future.

    These future objects must be awaited in the same task that created them.
    They cannot be used as input to other tasks.
    """

    def __init__(self, mailbox_id: int) -> None:
        """Initialize a future tied to a local mailbox."""

        self.mailbox_id = mailbox_id
        """The mailbox id where this future's result will be shipped."""

        self._next_flag = False
        """If flag is set, then awaiting returns next result."""

    def __await__(self) -> Any:
        """
        Wait on a result to be delivered.

        Informs the event loop which mailbox this is waiting on.
        """
        if self._next_flag:
            return (yield self)

        return (yield self)

    def __getstate__(self) -> Any:
        """Prevent a future from being sent to another process."""
        raise NotImplementedError(
            'These future objects must be awaited in the same task that'
            ' created them. They cannot be inputted to other tasks.',
        )

    @property
    def _done(self) -> bool:
        """
        Return true if the future is ready.

        Warning: Polling on futures in a busy-wait loop can cause deadlock.
        It is best to await on futures rather than continuously polling.
        Use at your own risk.
        """
        from bqskit.runtime.worker import get_worker

        if self.mailbox_id not in get_worker()._mailboxes:
            return True

        return get_worker()._mailboxes[self.mailbox_id].ready



"""This module implements the PAMLayoutPass."""
from __future__ import annotations

import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.point import CircuitPoint
from bqskit.passes.control.foreach import ForEachBlockPass
from bqskit.passes.mapping.pam import PAMBlockTAPermData
from bqskit.passes.mapping.pam import PermutationAwareMappingAlgorithm

_logger = logging.getLogger(__name__)


class PAMLayoutPass(PermutationAwareMappingAlgorithm, BasePass):
    """Layout algorithm using permutation-aware mapping."""

    def __init__(
        self,
        total_passes: int = 1,
        gate_count_weight: float = 0.3,
        decay_delta: float = 0.001,
        decay_reset_interval: int = 5,
        decay_reset_on_gate: bool = True,
        extended_set_size: int = 20,
        extended_set_weight: float = 0.5,
    ) -> None:
        """
        Construct a PAMLayoutPass.

        Args:
            total_passes (int): The amount of forward and backward passes
                to apply before finalizing the layout.

            gate_count_weight (float): See
                :class:`PermutationAwareMappingAlgorithm` for info.
                (Default: 0.3)

            decay_delta (float): See :class:`GeneralizedSabreAlgorithm`
                for info. (Default: 0.001)

            decay_reset_interval (int): See :class:`GeneralizedSabreAlgorithm`
                for info. (Default: 5)

            decay_reset_on_gate (bool): See :class:`GeneralizedSabreAlgorithm`
                for info. (Default: True)

            extended_set_size (int): See :class:`GeneralizedSabreAlgorithm`
                for info. (Default: 20)

            extended_set_weight (float): See :class:`GeneralizedSabreAlgorithm`
                for info. (Default: 0.5)
        """
        if not isinstance(total_passes, int):
            m = f'Expected int for total_passes, got {type(total_passes)}'
            raise TypeError(m)

        if total_passes < 1:
            raise ValueError('Total passes must be a positive integer.')

        self.total_passes = total_passes
        super().__init__(
            gate_count_weight,
            decay_delta,
            decay_reset_interval,
            decay_reset_on_gate,
            extended_set_size,
            extended_set_weight,
        )

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        subgraph = data.connectivity

        perm_data: dict[CircuitPoint, PAMBlockTAPermData] = {}
        block_datas = data[ForEachBlockPass.key][-1]
        for block_data in block_datas:
            perm_data[block_data['point']] = block_data['permutation_data']

        if not subgraph.is_fully_connected():
            raise RuntimeError('Cannot route circuit on disconnected qudits.')

        pi = [i for i in range(circuit.num_qudits)]

        for _ in range(self.total_passes):
            self.forward_pass(circuit, pi, subgraph, perm_data)
            self.backward_pass(circuit, pi, subgraph)

        self._apply_perm(pi, data.placement)
        _logger.info(f'Found layout: {pi}, new placement: {data.placement}')



"""This module implements the SetRandomSeedPass class."""
from __future__ import annotations

import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.utils.typing import is_integer


_logger = logging.getLogger(__name__)


class SetRandomSeedPass(BasePass):
    """
    The SetRandomSeedPass class.

    The SetRandomSeedPass sets the random seed.
    """

    def __init__(self, seed: int = 0) -> None:
        """
        Construct a SetRandomSeedPass.

        Args:
            seed (int): The value to set the random seed to.
        """

        if not is_integer(seed):
            raise TypeError('Expected integer seed, got %s.' % type(seed))

        self.seed = seed

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        _logger.info(f'Setting the random seed to {self.seed}.')
        data.seed = self.seed



"""This module implemenets Honeywell QPU models."""
from __future__ import annotations

import warnings

from bqskit.compiler.machine import MachineModel
from bqskit.ir.gate import Gate
from bqskit.ir.gates.constant.zz import ZZGate
from bqskit.ir.gates.parameterized.rz import RZGate
from bqskit.ir.gates.parameterized.u1q import U1qPi2Gate
from bqskit.ir.gates.parameterized.u1q import U1qPiGate

warnings.warn(
    'Honeywell Quantum is now Quantinuum. Please use the '
    'Quantinuum QPU models and gate sets instead. This warning will become'
    'an error in a future version of BQSKit.',
    DeprecationWarning,
)

honeywell_gate_set: set[Gate] = {U1qPiGate, U1qPi2Gate, RZGate(), ZZGate()}

H1_1Model = MachineModel(20, None, honeywell_gate_set)
H1_2Model = MachineModel(12, None, honeywell_gate_set)



"""This package contains integrations with the QuTiP framework."""
from __future__ import annotations

from bqskit.ext.qutip.translate import bqskit_to_qutip
from bqskit.ext.qutip.translate import qutip_to_bqskit

__all__ = ['bqskit_to_qutip', 'qutip_to_bqskit']



"""This module implements the Unitary abstract base class."""
from __future__ import annotations

import abc
from typing import Sequence
from typing import TYPE_CHECKING
from typing import Union

import numpy as np

from bqskit.qis.unitary.meta import UnitaryMeta
from bqskit.utils.typing import is_real_number
from bqskit.utils.typing import is_sequence

if TYPE_CHECKING:
    from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class Unitary(metaclass=UnitaryMeta):
    """
    A unitary-valued function.

    A `Unitary` is a map from zero or more real numbers to a unitary matrix.
    This is captured in the `get_unitary` abstract method.
    """

    _num_params: int
    _num_qudits: int
    _radixes: tuple[int, ...]
    _dim: int

    @property
    def num_params(self) -> int:
        """The number of real parameters this unitary-valued function takes."""
        return getattr(self, '_num_params')

    @property
    def num_qudits(self) -> int:
        """The number of qudits this unitary can act on."""
        if hasattr(self, '_num_qudits'):
            return self._num_qudits

        return len(self.radixes)

    @property
    def radixes(self) -> tuple[int, ...]:
        """The number of orthogonal states for each qudit."""
        return getattr(self, '_radixes')

    @property
    def dim(self) -> int:
        """The matrix dimension for this unitary."""
        if hasattr(self, '_dim'):
            return self._dim

        return int(np.prod(self.radixes))

    @abc.abstractmethod
    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """
        Map real-valued `params` to a `UnitaryMatrix`.

        Args:
            params (RealVector): Unconstrained vector of real number
                parameters for parameterized unitaries.

        Returns:
            UnitaryMatrix: The unitary matrix.
        """

    def is_qubit_only(self) -> bool:
        """Return true if this unitary can only act on qubits."""
        return all([radix == 2 for radix in self.radixes])

    def is_qutrit_only(self) -> bool:
        """Return true if this unitary can only act on qutrits."""
        return all([radix == 3 for radix in self.radixes])

    def is_qudit_only(self, radix: int) -> bool:
        """
        Return true if this unitary can only act on `radix`-qudits.

        Args:
            radix (int): Check all qudits have this many orthogonal
                states.
        """
        return all([r == radix for r in self.radixes])

    def is_parameterized(self) -> bool:
        """Return true if this unitary is parameterized."""
        return self.num_params != 0

    def is_constant(self) -> bool:
        """Return true if this unitary doesn't take parameters."""
        return not self.is_parameterized()

    def check_parameters(self, params: RealVector) -> None:
        """
        Check parameters are valid and match the unitary.

        Args:
            params(RealVector): The parameters to check.

        Raises:
            ValueError: If parameter length does not match expected number.
        """
        if not is_sequence(params):
            raise TypeError(
                'Expected a sequence type for params, got %s.'
                % type(params),
            )

        if not all(is_real_number(p) for p in params):
            typechecks = [is_real_number(p) for p in params]
            fail_idx = typechecks.index(False)
            raise TypeError(
                'Expected params to be floats, got %s.'
                % type(params[fail_idx]),
            )

        if len(params) != self.num_params:
            raise ValueError(
                'Expected %d params, got %d.'
                % (self.num_params, len(params)),
            )

    def is_self_inverse(self, params: RealVector = []) -> bool:
        """
        Checks whether the unitary is its own inverse.

        A unitary is its own inverse if its matrix is equal to its
        Hermitian conjugate.

        Args:
            params (RealVector): The parameters of the unitary to check.

        Returns:
            bool: True if the unitary is self-inverse, False otherwise.

        Note:
            - This checks that the unitary is self-inverse for the given
              parameters only.
        """
        # Get the unitary matrix of the gate
        unitary_matrix = self.get_unitary(params)

        # Calculate the Hermitian conjugate (adjoint) of the unitary matrix
        hermitian_conjugate = unitary_matrix.dagger

        # Check if the unitary matrix is equal to its Hermitian conjugate
        return np.allclose(unitary_matrix, hermitian_conjugate)


RealVector = Union[Sequence[float], np.ndarray]



"""This module implements the MeasurementPlaceholder class."""
from __future__ import annotations

from bqskit.ir.gate import Gate
from bqskit.ir.location import CircuitLocation
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class MeasurementPlaceholder(Gate):
    """Pseudogate to hold measurement information."""

    def __init__(
        self,
        classical_regs: list[tuple[str, int]],
        measurements: dict[int, tuple[str, int]],
    ) -> None:
        """
        Construct a MeasurementPlaceholder.

        Args:
            classical_regs (list[tuple[str, int]]): A list of classical
                register descriptors. Each one is given as a tuple containing
                its name and size.

            measurements (dict[int, tuple[str, int]]): A list of measurements
                made. Measurements are given as a map of qudit index to
                a tuple containing the classical register's name and index.
        """
        self._name = 'measurement'
        self._num_qudits = len(measurements)
        self._radixes = tuple([2] * self._num_qudits)
        self._num_params = 0
        self.classical_regs = classical_regs
        self.measurements = measurements

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        raise RuntimeError(
            'Cannot compute unitary for a measurement placeholder.'
            ' Consider removing measurements either by calling'
            ' `circuit.remove_measurements()` or by using the'
            ' ExtractMeasurements and RestoreMeasurements passes.',
        )

    def get_qasm_gate_def(self) -> str:
        """Declares the classical registers."""
        ret = ''
        for name, size in self.classical_regs:
            ret += f'creg {name}[{size}];\n'
        return ret

    def get_qasm(self, params: RealVector, location: CircuitLocation) -> str:
        ret = ''
        for qudit, (name, index) in self.measurements.items():
            ret += f'measure q[{qudit}] -> {name}[{index}];\n'
        return ret

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, MeasurementPlaceholder)
            and other.measurements == self.measurements
            and other.classical_regs == self.classical_regs
        )

    def __hash__(self) -> int:
        return hash(tuple(self.classical_regs))



"""This module implements the UnitaryMeta metaclass."""
from __future__ import annotations

import abc
from typing import Any


class UnitaryMeta(abc.ABCMeta):
    """
    The UnitaryMeta metaclass.

    Necessary to provide isinstance checks for composed classes.
    """

    def __instancecheck__(cls, instance: Any) -> bool:
        """
        Check if an instance is a `Unitary` instance.

        Additional checks for DifferentiableUnitary and
        LocallyOptimizableUnitary. We check if the object has the
        is_differentiable or is_locally_optimizable callable, an instance method
        that maps nothing to a bool. If the object has the method, then it must
        return true for isinstance to pass.

        This can be used with composed classes to implement conditional
        inheritance.
        """
        if cls.__name__ == 'DifferentiableUnitary':
            if hasattr(instance, 'is_differentiable'):
                if not instance.is_differentiable():
                    return False

        if cls.__name__ == 'LocallyOptimizableUnitary':
            if hasattr(instance, 'is_locally_optimizable'):
                if not instance.is_locally_optimizable():
                    return False

        return super().__instancecheck__(instance)



"""This module implements the GeneralizedSabreLayoutPass class."""
from __future__ import annotations

import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.passes.mapping.sabre import GeneralizedSabreAlgorithm

_logger = logging.getLogger(__name__)


class GeneralizedSabreLayoutPass(BasePass, GeneralizedSabreAlgorithm):
    """
    Uses the Sabre algorithm to layout the circuit.

    See :class:`GeneralizedSabreAlgorithm` for more info.
    """

    def __init__(
        self,
        total_passes: int = 1,
        decay_delta: float = 0.001,
        decay_reset_interval: int = 5,
        decay_reset_on_gate: bool = True,
        extended_set_size: int = 20,
        extended_set_weight: float = 0.5,
    ) -> None:
        """
        Construct a GeneralizedSabreLayoutPass.

        Args:
            total_passes (int): The amount of forward and backward passes
                to apply before finalizing the layout.

            decay_delta (float): See :class:`GeneralizedSabreAlgorithm`
                for info. (Default: 0.001)

            decay_reset_interval (int): See :class:`GeneralizedSabreAlgorithm`
                for info. (Default: 5)

            decay_reset_on_gate (bool): See :class:`GeneralizedSabreAlgorithm`
                for info. (Default: True)

            extended_set_size (int): See :class:`GeneralizedSabreAlgorithm`
                for info. (Default: 20)

            extended_set_weight (float): See :class:`GeneralizedSabreAlgorithm`
                for info. (Default: 0.5)
        """
        if not isinstance(total_passes, int):
            raise TypeError(
                'Expected int for total_passes'
                f', got {type(total_passes)}',
            )

        if total_passes < 1:
            raise ValueError('Total passes must be a positive integer.')

        self.total_passes = total_passes
        super().__init__(
            decay_delta,
            decay_reset_interval,
            decay_reset_on_gate,
            extended_set_size,
            extended_set_weight,
        )

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        subgraph = data.connectivity
        if not subgraph.is_fully_connected():
            raise RuntimeError('Cannot layout circuit on disconnected qudits.')

        pi = [i for i in range(circuit.num_qudits)]

        for _ in range(self.total_passes):
            self.forward_pass(circuit, pi, subgraph)
            self.backward_pass(circuit, pi, subgraph)

        self._apply_perm(pi, data.placement)
        _logger.info(f'Found layout: {pi}, new placement: {data.placement}')



"""This module implements the BarrierPlaceholder class."""
from __future__ import annotations

from typing import Sequence

from bqskit.ir.gate import Gate
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class BarrierPlaceholder(Gate):
    """Pseudogate to force break partitioning algorithms."""

    def __init__(self, num_qudits: int, radixes: Sequence[int] = []) -> None:
        """Construct a BarrierPlaceholder."""
        self._name = 'barrier'
        self._qasm_name = 'barrier'
        self._num_qudits = num_qudits
        self._radixes = tuple(radixes) if radixes else tuple([2] * num_qudits)
        self._num_params = 0

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        return UnitaryMatrix.identity(self.dim, self.radixes)

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, BarrierPlaceholder)
            and other.num_qudits == self.num_qudits
            and other.radixes == self.radixes
        )

    def __hash__(self) -> int:
        return hash(('BarrierPlaceholder', self.num_qudits, self.radixes))



"""This module implements the IfThenElsePass class."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.workflow import Workflow
from bqskit.passes.control.predicate import PassPredicate

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit
    from bqskit.compiler.passdata import PassData
    from bqskit.compiler.workflow import WorkflowLike

_logger = logging.getLogger(__name__)


class IfThenElsePass(BasePass):
    """
    The IfThenElsePass class.

    This is a control pass that conditionally executes a workflow.
    """

    def __init__(
        self,
        condition: PassPredicate,
        on_true: WorkflowLike,
        on_false: WorkflowLike | None = None,
    ) -> None:
        """
        Construct a IfThenElsePass.

        Args:
            condition (PassPredicate): The condition checked.

            on_true (WorkflowLike): The pass or passes to execute if
                `condition` is true.

            on_false (WorkflowLike | None): If specified,
                the pass or passes to execute if `condition` is false.
                Defaults to None, which does is equivalent to a No-Op.
        """

        if not isinstance(condition, PassPredicate):
            raise TypeError('Expected PassPredicate, got %s.' % type(condition))

        self.condition = condition
        self.on_true = Workflow(on_true)
        self.on_false = Workflow(on_false) if on_false is not None else None

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        if self.condition(circuit, data):
            _logger.debug('True branch taken.')
            await self.on_true.run(circuit, data)

        elif self.on_false is not None:
            _logger.debug('False branch taken.')
            await self.on_false.run(circuit, data)



"""
This package implements most OPENQASM 2.0 reading and writing features.

References:
    Andrew W. Cross, Lev S. Bishop, John A. Smolin, Jay M. Gambetta
    "Open Quantum Assembly Language" [arXiv:1707.03429].
"""
from __future__ import annotations

from bqskit.ir.lang.qasm2.qasm2 import OPENQASM2Language

__all__ = ['OPENQASM2Language']



"""
=============================================================
Circuit Execution Module (:mod:`bqskit.exec`)
=============================================================

.. currentmodule:: bqskit.exec

BQSKit supports a variety of ways to execute circuits through
CircuitRunner.

.. rubric:: Core Classes

.. autosummary::
    :toctree: autogen
    :recursive:

    CircuitRunner
    RunnerResults

.. rubric:: CircuitRunners

.. autosummary::
    :toctree: autogen
    :recursive:

    QuestRunner
    IBMQRunner
    SimulationRunner
"""
from __future__ import annotations

__all__ = [
    'CircuitRunner',
    'RunnerResults',
    'QuestRunner',
    'SimulationRunner',
    'gen_approximate_circuits',
]

from bqskit.exec.results import RunnerResults
from bqskit.exec.runner import CircuitRunner
try:
    from bqskit.exec.runners.ibmq import IBMQRunner
    __all__ += ['IBMQRunner']
except ImportError:
    pass
from bqskit.exec.runners.quest import QuestRunner
from bqskit.exec.runners.sim import SimulationRunner
from bqskit.exec.runners.quest import gen_approximate_circuits



"""This module implements the YGate."""
from __future__ import annotations

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class YGate(ConstantGate, QubitGate):
    """
    The Pauli Y gate.

    The Y gate is given by the following unitary:

    .. math::

        \\begin{pmatrix}
        0 & -i \\\\
        i & 0 \\\\
        \\end{pmatrix}
    """

    _num_qudits = 1
    _qasm_name = 'y'
    _utry = UnitaryMatrix(
        [
            [0, -1j],
            [1j, 0],
        ],
    )



"""
=======================================================
A Runtime for the Compile-time (:mod:`bqskit.runtime`)
=======================================================

.. currentmodule:: bqskit.runtime

This package provides an execution environment for compiler passes,
enabling them to parallelize and distribute their workload efficiently.

Launching a Server
##################

To parallelize BQSKit compilations, you can launch a BQSKit Runtime
server in either attached or detached mode, and submit a
:class:`~bqskit.compiler.task.CompilationTask` through a connected
:class:`~bqskit.compiler.compiler.Compiler`.

By default, a compiler will automatically start and connect to an attached
runtime server during initialization. In this mode, the compiler
starts up and shuts down the server automatically and transparently to
the end user. See :class:`~bqskit.runtime.attached.AttachedServer` for
more info.

A server running in attached mode is limited to a single machine. If you
would like to parallelize BQSKit across multiple nodes over a network,
you can launch a server independently in detached mode. Here, you first
start :class:`~bqskit.runtime.manager.Manager` processes on all machines.
Then link them all together by running a
:class:`~bqskit.runtime.detached.DetachedServer` process, which Compilers'
can connect to by passing the appropriate address in their constructor.

Upon installing BQSKit, two shell scripts are added to your environment
that can spawn managers and detached servers::

    bqskit-manager

and::

    bqskit-server <address_of_bqskit_managers>

Typically, you start managers first on all nodes in the desired cluster.
Then you can start the server with a comma-separated list of all managers
ip address and optionally ports. Once a server is started and has connected
to the requested managers, no more managers can be added. You can see
the `-h` option of each command or the
:func:`~bqskit.runtime.detached.start_server` and
:func:`~bqskit.runtime.manager.start_manager` entry points for
more information.

For more information on how to manage jobs submitted to a runtime server,
see the :obj:`~bqskit.compiler` documentation.

.. autosummary::
    :toctree: autogen
    :recursive:

    ~bqskit.runtime.attached.AttachedServer
    ~bqskit.runtime.detached.DetachedServer
    ~bqskit.runtime.manager.Manager
    ~bqskit.runtime.detached.start_server
    ~bqskit.runtime.manager.start_manager
    ~bqskit.runtime.worker.start_worker_rank

Parallelize Pass Computation
############################

When developing a BQSKit compiler pass, you can hook into the active
runtime server through the :func:`get_runtime` function. This returns a
:class:`RuntimeHandle`, which you can use to submit, map, wait on, and
cancel tasks in the execution environment.

For more information on how to design a custom pass, see this (TODO, sorry,
you can look at the source code of existing
`passes <https://github.com/BQSKit/bqskit/blob/master/bqskit/passes>`_
for a good example for the time being).

.. autosummary::
    :toctree: autogen
    :recursive:

    ~bqskit.runtime.get_runtime
    ~bqskit.runtime.RuntimeHandle
    ~bqskit.runtime.future.RuntimeFuture

Standard Logging Interface
##########################

This system uses the standard python logging module to provide a familiar
logging interface. On client processes -- the ones where you create the
Compiler object -- you can configure python loggers like normal before
submitting tasks and have the entire system honor that configuration.
"""
from __future__ import annotations

import faulthandler
import os
from typing import Any
from typing import Callable
from typing import Protocol
from typing import TYPE_CHECKING

# Enable low-level fault handling: system crashes print a minimal trace.
faulthandler.enable()


# Disable multi-threading in BLAS libraries.
os.environ['OMP_NUM_THREADS'] = '1'
os.environ['OPENBLAS_NUM_THREADS'] = '1'
os.environ['MKL_NUM_THREADS'] = '1'
os.environ['NUMEXPR_NUM_THREADS'] = '1'
os.environ['VECLIB_MAXIMUM_THREADS'] = '1'
os.environ['RUST_BACKTRACE'] = '1'


if TYPE_CHECKING:
    from bqskit.runtime.future import RuntimeFuture


class RuntimeHandle(Protocol):
    """
    A structural type capturing BQSKit Runtime capabilities.

    A compiler pass designer can effortlessly parallelize certain aspects
    of compilation using the :func:`submit` and :func:`map` methods.
    These both will return :class:`~bqskit.runtime.future.RuntimeFuture`
    objects, which can be awaited on using the standard python `await`
    keyword. This can also be used to cancel tasks and wait on specific
    events.

    This should never be constructed directly and only accessed through
    the :func:`get_runtime` function, since this is a structural type
    definition and does not contain any implementation.
    """

    def submit(
        self,
        fn: Callable[..., Any],
        *args: Any,
        **kwargs: Any,
    ) -> RuntimeFuture:
        """Submit a `fn` to the runtime."""
        ...

    def map(
        self,
        fn: Callable[..., Any],
        *args: Any,
        **kwargs: Any,
    ) -> RuntimeFuture:
        """Map `fn` over the input arguments distributed across the runtime."""
        ...

    def cancel(self, future: RuntimeFuture) -> None:
        """
        Cancel all tasks associated with `future`.

        This is asynchronous and non-pre-emptive. The cancel will return
        immediately, the future it represents will become dead, and at some
        point in the future, all tasks and data associated with the cancelled
        one will be removed from the system.
        """
        ...

    async def next(self, future: RuntimeFuture) -> list[tuple[int, Any]]:
        """
        Wait for and return the next batch of results from a map task.

        Returns:
            (list[tuple[int, Any]]): A list of the results that arrived
                while the task was waiting. Each result is paired with
                the index of its arguments in the original map call.
        """
        ...


def get_runtime() -> RuntimeHandle:
    """
    Return a handle on the active runtime.

    See :class:`RuntimeHandle` for more info.
    """
    from bqskit.runtime.worker import get_worker
    return get_worker()


default_server_port = 7472
default_manager_port = 7473
default_worker_port = 7474
__all__ = ['get_runtime']



"""
====================================================================
Compilation and Synthesis Algorithms (:mod:`bqskit.passes`)
====================================================================

.. rubric:: Partitioning Passes

These passes will group together gates in the circuit. It is recommended
to use the QuickPartitioner currently.

.. autosummary::
    :toctree: autogen
    :recursive:

    ClusteringPartitioner
    GreedyPartitioner
    ScanPartitioner
    QuickPartitioner
    GroupSingleQuditGatePass

.. rubric:: Synthesis Passes

These passes perform synthesis. Decomposition passes break down
unitaries into smaller ones and will need to be followed up by another
synthesis pass to convert the circuit to native gates.

.. autosummary::
    :toctree: autogen
    :recursive:

    LEAPSynthesisPass
    QSearchSynthesisPass
    QFASTDecompositionPass
    QPredictDecompositionPass
    SynthesisPass

.. rubric:: Processing Passes

.. autosummary::
    :toctree: autogen
    :recursive:

    ExhaustiveGateRemovalPass
    IterativeScanningGateRemovalPass
    ScanningGateRemovalPass
    SubstitutePass

.. rubric:: Retargeting Passes

.. autosummary::
    :toctree: autogen
    :recursive:

    AutoRebase2QuditGatePass
    GeneralSQDecomposition
    Rebase2QuditGatePass

.. rubric:: Control Passes

This are passes that are used to manage control flow inside of a workflow.

.. autosummary::
    :toctree: autogen
    :recursive:

    DoWhileLoopPass
    ForEachBlockPass
    IfThenElsePass
    WhileLoopPass
    DoThenDecide
    ParallelDo

.. rubric:: Predicates

This objects are designed as conditions for use with control passes.

.. autosummary::
    :toctree: autogen
    :recursive:

    PassPredicate
    ChangePredicate
    GateCountPredicate
    ManyQuditGatesPredicate
    NotPredicate
    WidthPredicate
    PhysicalPredicate
    SinglePhysicalPredicate
    MultiPhysicalPredicate
    NoSingleQuditGatesInModel
    HasGeneralSingleQuditGate
    ZXGatePredicate
    AllConstantSingleQuditGates

.. rubric:: Rule-based Passes

These passes apply fixed rules to a circuit to transform it.

.. autosummary::
    :toctree: autogen
    :recursive:

    CHToCNOTPass
    CNOTToCZPass
    CNOTToCHPass
    CNOTToCYPass
    CYToCNOTPass
    SwapToCNOTPass
    U3Decomposition
    ZXZXZDecomposition

.. rubric:: Mapping Passes

These passes either perform qubit placement, layout, routing or otherwise
are involved the qubit mapping process.

.. autosummary::
    :toctree: autogen
    :recursive:

    GeneralizedSabreLayoutPass
    GreedyPlacementPass
    TrivialPlacementPass
    GeneralizedSabreRoutingPass
    SetModelPass
    ApplyPlacement
    SubtopologySelectionPass
    PAMLayoutPass
    PAMRoutingPass
    EmbedAllPermutationsPass
    ExtractModelConnectivityPass
    RestoreModelConnevtivityPass


.. rubric:: PAM Verification Passes

These passes either perform upper-bound error analysis of the PAM process.

    TagPAMBlockDataPass
    CalculatePAMErrorsPass
    UnTagPAMBlockDataPass
    PAMVerificationSequence

.. rubric:: Utility Passes

.. autosummary::
    :toctree: autogen
    :recursive:

    CompressPass
    RecordStatsPass
    SetRandomSeedPass
    UnfoldPass
    UpdateDataPass
    ToU3Pass
    ToVariablePass
    BlockConversionPass
    LogPass
    ExtendBlockSizePass
    LogErrorPass
    FillSingleQuditGatesPass
    StructureAnalysisPass
    ClearAllBlockData

.. rubric:: IO Passes

.. autosummary::
    :toctree: autogen
    :recursive:

    LoadCheckpointPass
    SaveCheckpointPass
    SaveIntermediatePass
    RestoreIntermediatePass

.. rubric:: Search Heuristics

.. autosummary::
    :toctree: autogen
    :recursive:

    HeuristicFunction
    AStarHeuristic
    GreedyHeuristic
    DijkstraHeuristic

.. rubric:: Search Layer Generators

Layer generators can be used to modify how search based synthesis
algorithms extend circuit templates.

.. autosummary::
    :toctree: autogen
    :recursive:

    FourParamGenerator
    MiddleOutLayerGenerator
    SeedLayerGenerator
    SimpleLayerGenerator
    SingleQuditLayerGenerator
    StairLayerGenerator
    WideLayerGenerator
"""
from __future__ import annotations

from bqskit.passes.alias import PassAlias
from bqskit.passes.control.dothendecide import DoThenDecide
from bqskit.passes.control.dowhileloop import DoWhileLoopPass
from bqskit.passes.control.foreach import ClearAllBlockData
from bqskit.passes.control.foreach import ForEachBlockPass
from bqskit.passes.control.ifthenelse import IfThenElsePass
from bqskit.passes.control.paralleldo import ParallelDo
from bqskit.passes.control.predicate import PassPredicate
from bqskit.passes.control.predicates.change import ChangePredicate
from bqskit.passes.control.predicates.count import GateCountPredicate
from bqskit.passes.control.predicates.many import ManyQuditGatesPredicate
from bqskit.passes.control.predicates.multi import MultiPhysicalPredicate
from bqskit.passes.control.predicates.notpredicate import NotPredicate
from bqskit.passes.control.predicates.physical import PhysicalPredicate
from bqskit.passes.control.predicates.single import AllConstantSingleQuditGates
from bqskit.passes.control.predicates.single import HasGeneralSingleQuditGate
from bqskit.passes.control.predicates.single import NoSingleQuditGatesInModel
from bqskit.passes.control.predicates.single import SinglePhysicalPredicate
from bqskit.passes.control.predicates.single import ZXGatePredicate
from bqskit.passes.control.predicates.width import WidthPredicate
from bqskit.passes.control.whileloop import WhileLoopPass
from bqskit.passes.group import PassGroup
from bqskit.passes.io.checkpoint import LoadCheckpointPass
from bqskit.passes.io.checkpoint import SaveCheckpointPass
from bqskit.passes.io.intermediate import RestoreIntermediatePass
from bqskit.passes.io.intermediate import SaveIntermediatePass
from bqskit.passes.mapping.apply import ApplyPlacement
from bqskit.passes.mapping.embed import EmbedAllPermutationsPass
from bqskit.passes.mapping.layout.pam import PAMLayoutPass
from bqskit.passes.mapping.layout.sabre import GeneralizedSabreLayoutPass
from bqskit.passes.mapping.placement.greedy import GreedyPlacementPass
from bqskit.passes.mapping.placement.trivial import TrivialPlacementPass
from bqskit.passes.mapping.routing.pam import PAMRoutingPass
from bqskit.passes.mapping.routing.sabre import GeneralizedSabreRoutingPass
from bqskit.passes.mapping.setmodel import ExtractModelConnectivityPass
from bqskit.passes.mapping.setmodel import RestoreModelConnevtivityPass
from bqskit.passes.mapping.setmodel import SetModelPass
from bqskit.passes.mapping.topology import SubtopologySelectionPass
from bqskit.passes.mapping.verify import CalculatePAMErrorsPass
from bqskit.passes.mapping.verify import PAMVerificationSequence
from bqskit.passes.mapping.verify import TagPAMBlockDataPass
from bqskit.passes.mapping.verify import UnTagPAMBlockDataPass
from bqskit.passes.measure import ExtractMeasurements
from bqskit.passes.measure import RestoreMeasurements
from bqskit.passes.noop import NOOPPass
from bqskit.passes.partitioning.cluster import ClusteringPartitioner
from bqskit.passes.partitioning.greedy import GreedyPartitioner
from bqskit.passes.partitioning.quick import QuickPartitioner
from bqskit.passes.partitioning.scan import ScanPartitioner
from bqskit.passes.partitioning.single import GroupSingleQuditGatePass
from bqskit.passes.processing.exhaustive import ExhaustiveGateRemovalPass
from bqskit.passes.processing.iterative import IterativeScanningGateRemovalPass
from bqskit.passes.processing.scan import ScanningGateRemovalPass
from bqskit.passes.processing.substitute import SubstitutePass
from bqskit.passes.retarget.auto import AutoRebase2QuditGatePass
from bqskit.passes.retarget.general import GeneralSQDecomposition
from bqskit.passes.retarget.two import Rebase2QuditGatePass
from bqskit.passes.rules.ch2cnot import CHToCNOTPass
from bqskit.passes.rules.cnot2ch import CNOTToCHPass
from bqskit.passes.rules.cnot2cy import CNOTToCYPass
from bqskit.passes.rules.cnot2cz import CNOTToCZPass
from bqskit.passes.rules.cy2cnot import CYToCNOTPass
from bqskit.passes.rules.swap2cnot import SwapToCNOTPass
from bqskit.passes.rules.u3 import U3Decomposition
from bqskit.passes.rules.zxzxz import ZXZXZDecomposition
from bqskit.passes.search.frontier import Frontier
from bqskit.passes.search.generator import LayerGenerator
from bqskit.passes.search.generators.fourparam import FourParamGenerator
from bqskit.passes.search.generators.middleout import MiddleOutLayerGenerator
from bqskit.passes.search.generators.seed import SeedLayerGenerator
from bqskit.passes.search.generators.simple import SimpleLayerGenerator
from bqskit.passes.search.generators.single import SingleQuditLayerGenerator
from bqskit.passes.search.generators.stair import StairLayerGenerator
from bqskit.passes.search.generators.wide import WideLayerGenerator
from bqskit.passes.search.heuristic import HeuristicFunction
from bqskit.passes.search.heuristics.astar import AStarHeuristic
from bqskit.passes.search.heuristics.dijkstra import DijkstraHeuristic
from bqskit.passes.search.heuristics.greedy import GreedyHeuristic
from bqskit.passes.synthesis.leap import LEAPSynthesisPass
from bqskit.passes.synthesis.pas import PermutationAwareSynthesisPass
from bqskit.passes.synthesis.qfast import QFASTDecompositionPass
from bqskit.passes.synthesis.qpredict import QPredictDecompositionPass
from bqskit.passes.synthesis.qsearch import QSearchSynthesisPass
from bqskit.passes.synthesis.synthesis import SynthesisPass
from bqskit.passes.synthesis.target import SetTargetPass
from bqskit.passes.util.compress import CompressPass
from bqskit.passes.util.conversion import BlockConversionPass
from bqskit.passes.util.converttou3 import ToU3Pass
from bqskit.passes.util.converttovar import ToVariablePass
from bqskit.passes.util.extend import ExtendBlockSizePass
from bqskit.passes.util.fill import FillSingleQuditGatesPass
from bqskit.passes.util.log import LogErrorPass
from bqskit.passes.util.log import LogPass
from bqskit.passes.util.random import SetRandomSeedPass
from bqskit.passes.util.record import RecordStatsPass
from bqskit.passes.util.structure import StructureAnalysisPass
from bqskit.passes.util.unfold import UnfoldPass
from bqskit.passes.util.update import UpdateDataPass


__all__ = [
    'DoWhileLoopPass',
    'ClearAllBlockData',
    'ForEachBlockPass',
    'IfThenElsePass',
    'PassPredicate',
    'ChangePredicate',
    'GateCountPredicate',
    'NotPredicate',
    'WhileLoopPass',
    'ClusteringPartitioner',
    'GreedyPartitioner',
    'ScanPartitioner',
    'QuickPartitioner',
    'SynthesisPass',
    'LEAPSynthesisPass',
    'QSearchSynthesisPass',
    'QFASTDecompositionPass',
    'QPredictDecompositionPass',
    'CompressPass',
    'RecordStatsPass',
    'SetRandomSeedPass',
    'UnfoldPass',
    'UpdateDataPass',
    'ToU3Pass',
    'ScanningGateRemovalPass',
    'SimpleLayerGenerator',
    'AStarHeuristic',
    'GreedyHeuristic',
    'DijkstraHeuristic',
    'Frontier',
    'LayerGenerator',
    'HeuristicFunction',
    'SeedLayerGenerator',
    'BlockConversionPass',
    'StairLayerGenerator',
    'DoThenDecide',
    'SubstitutePass',
    'ParallelDo',
    'LoadCheckpointPass',
    'SaveCheckpointPass',
    'SaveIntermediatePass',
    'RestoreIntermediatePass',
    'GroupSingleQuditGatePass',
    'SingleQuditLayerGenerator',
    'MiddleOutLayerGenerator',
    'CNOTToCZPass',
    'ZXZXZDecomposition',
    'ExhaustiveGateRemovalPass',
    'WidthPredicate',
    'IterativeScanningGateRemovalPass',
    'SetModelPass',
    'PassAlias',
    'PassGroup',
    'NOOPPass',
    'FourParamGenerator',
    'WideLayerGenerator',
    'SwapToCNOTPass',
    'GeneralizedSabreLayoutPass',
    'GreedyPlacementPass',
    'TrivialPlacementPass',
    'GeneralizedSabreRoutingPass',
    'SetModelPass',
    'U3Decomposition',
    'PhysicalPredicate',
    'SinglePhysicalPredicate',
    'Rebase2QuditGatePass',
    'LogPass',
    'ExtractMeasurements',
    'RestoreMeasurements',
    'ExtendBlockSizePass',
    'ApplyPlacement',
    'MultiPhysicalPredicate',
    'LogErrorPass',
    'FillSingleQuditGatesPass',
    'CHToCNOTPass',
    'CNOTToCHPass',
    'CNOTToCYPass',
    'CYToCNOTPass',
    'SetTargetPass',
    'PAMLayoutPass',
    'PAMRoutingPass',
    'EmbedAllPermutationsPass',
    'SubtopologySelectionPass',
    'PermutationAwareSynthesisPass',
    'ToVariablePass',
    'AutoRebase2QuditGatePass',
    'ManyQuditGatesPredicate',
    'NoSingleQuditGatesInModel',
    'HasGeneralSingleQuditGate',
    'ZXGatePredicate',
    'AllConstantSingleQuditGates',
    'GeneralSQDecomposition',
    'StructureAnalysisPass',
    'ExtractModelConnectivityPass',
    'RestoreModelConnevtivityPass',
    'TagPAMBlockDataPass',
    'CalculatePAMErrorsPass',
    'UnTagPAMBlockDataPass',
    'PAMVerificationSequence',
]



"""This module implements the DoThenDecide class."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.workflow import Workflow

if TYPE_CHECKING:
    from typing import Callable

    from bqskit.compiler.passdata import PassData
    from bqskit.compiler.workflow import WorkflowLike
    from bqskit.ir.circuit import Circuit


_logger = logging.getLogger(__name__)


class DoThenDecide(BasePass):
    """
    The DoThenDecide class.

    This is a control pass that executes a workflow and then conditionally
    accepts the resulting circuit or reverts to the original state.
    """

    def __init__(
        self,
        condition: Callable[[Circuit, Circuit], bool],
        workflow: WorkflowLike,
    ) -> None:
        """
        Construct a DoThenDecide.

        Args:
            condition (Callable[[Circuit, Circuit], bool]): The condition
                function that determines if the new circuit
                (second parameter) after running `workflow` should replace
                the original circuit (first parameter). If the condition
                returns True, then replace the original circuit with the
                new one. Note when passing callables to BQSKit passes,
                they need to be defined at the module level (0-indent)
                with a name (no lambdas).

            workflow (WorkflowLike): The pass or passes to execute.
        """
        if not callable(condition):
            bad_type = type(condition)
            msg = f'Expected callable function for condition, got {bad_type}'
            raise TypeError(msg)

        self.condition = condition
        self.workflow = Workflow(workflow)

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        # Store old state
        old_circuit = circuit.copy()
        old_data = data.copy()

        # Execute the workflow
        await self.workflow.run(circuit, data)

        # Evaluate condition
        if self.condition(old_circuit, circuit):
            _logger.info('Accepted circuit.')

        else:
            circuit.become(old_circuit)
            data.become(old_data)
            _logger.info('Rejected circuit.')



"""This module implements numerical functions."""
from __future__ import annotations

from typing import Any

import numpy as np
import numpy.typing as npt
import scipy as sp

from bqskit.qis.pauli import PauliMatrices
from bqskit.qis.unitary.unitary import RealVector


def dexpmv(
    M: npt.NDArray[np.complex128], dM: npt.NDArray[np.complex128],
) -> tuple[npt.NDArray[np.complex128], npt.NDArray[np.complex128]]:
    """
    Compute the Matrix exponential F = e^M and its derivative dF.

    User must provide M and its derivative dM. If the argument dM is a
    vector of partials then dF will be the respective partial vector.
    This is done using a Pade Approximat with scaling and squaring.

    Args:
        M (np.ndarray): Matrix to exponentiate.

        dM (np.ndarray): Derivative(s) of M.

    Returns:
        tuple: Tuple containing
            - F (np.ndarray): Exponentiated matrix, i.e. e^M.

            - dF (np.ndarray): Derivative(s) of F.

    References:
        Brančík, Lubomír. "Matlab programs for matrix exponential function
        derivative evaluation." Proc. of Technical Computing Prague 2008
        (2008): 17-24.
    """

    norm = np.linalg.norm(M, np.inf)
    e = np.log2(norm) if norm != 0 else -np.inf
    r = int(max(0, e + 1))
    M = M / (2 ** r)
    dM = dM / (2 ** r)
    X = M
    Y = dM
    c = 0.5
    F = np.identity(M.shape[0]) + c * M
    D = np.identity(M.shape[0]) - c * M
    dF = c * dM
    dD = -c * dM
    q = 6
    p = True
    for k in range(2, q + 1):
        c = c * (q - k + 1) / (k * (2 * q - k + 1))
        Y = dM @ X + M @ Y
        X = M @ X
        cX = c * X
        cY = c * Y
        F = F + cX
        dF = dF + cY
        if p:
            D = D + cX
            dD = dD + cY
        else:
            D = D - cX
            dD = dD - cY
        p = not p
    Dinv = np.linalg.inv(D)
    F = Dinv @ F
    dF = Dinv @ (dF - dD @ F)

    for k in range(1, r + 1):
        dF = dF @ F + F @ dF
        F = F @ F

    return F, dF


def softmax(
    x: npt.NDArray[np.float64],
    beta: int = 20,
) -> npt.NDArray[np.float64]:
    """
    Computes the softmax of vector x.

    Args:
        x (np.ndarray): Input vector to softmax.

        beta (int): Beta coefficient to scale steepness of softmax.

    Returns:
        np.ndarray: Output vector of softmax.
    """

    shiftx = beta * (x - np.max(x))
    exps = np.exp(shiftx)
    return exps / np.sum(exps)


def dot_product(alpha: RealVector, sigma: RealVector) -> npt.NDArray[Any]:
    """
    Computes the standard dot product of `alpha` with `sigma`.

    Args:
        alpha (RealVector): The alpha vector.

        sigma (RealVector): The sigma vector.

    Returns:
        np.ndarray: Sum of element-wise multiplication of `alpha`
        and `sigma`.
    """

    return np.array(np.sum([a * s for a, s in zip(alpha, sigma)], 0))


def unitary_log_no_i(
        U: npt.NDArray[np.complex128],
) -> npt.NDArray[np.complex128]:
    """
    Solves for H in U = e^{iH}

    Args:
        U (np.ndarray): The unitary to decompose.

    Returns:
        np.ndarray: H in e^{iH} = U.

    Note:
        This assumes the input is unitary but does not check. The output
        is undefined on non-unitary inputs.
    """

    T, Z = sp.linalg.schur(U)
    T = np.diag(T)
    D = T / np.abs(T)
    D = np.diag(np.log(D))
    H0 = -1j * (Z @ D @ Z.conj().T)
    return 0.5 * H0 + 0.5 * H0.conj().T


def pauli_expansion(H: npt.NDArray[np.complex128]) -> npt.NDArray[np.float64]:
    """
    Computes a Pauli expansion of the hermitian matrix H.

    Args:
        H (np.ndarray): The hermitian matrix to expand.

    Returns:
        np.ndarray: The coefficients of a Pauli expansion for H,
        i.e., X dot Sigma = H where Sigma is Pauli matrices of
        same size of H.

    Note:
        This assumes the input is hermitian but does not check. The
        output is undefined on non-hermitian inputs.
    """

    # Change basis of H to Pauli Basis (solve for coefficients -> X)
    n = int(np.log2(len(H)))
    paulis = PauliMatrices(n)
    flatten_paulis = [np.reshape(pauli, 4 ** n) for pauli in paulis]
    flatten_H = np.reshape(H, 4 ** n)
    A = np.stack(flatten_paulis, axis=-1)
    X = np.real(np.matmul(np.linalg.inv(A), flatten_H))
    return np.array(X)


def compute_su_generators(n: int) -> npt.NDArray[np.complex128]:
    """
    Computes the Lie algebra generators for SU(n).

    Args:
        n (int): dimension of SU(n) algebra

    Returns:
        npt.NDArray[np.complex128]: list of the SU(N) generators,
        note that they are Hermitian, but not neccesarily unitary.

    Raises:
        ValueError: if n<=0

    References:
        https://walterpfeifer.ch/liealgebra/LieAlg_wieBuch4.pdf
    """
    # TODO HermitianMatrix objects

    if n <= 0:
        raise ValueError(f'Expected positive integer for n, got: {n}.')

    elif n == 1:
        return np.array([1], dtype=np.complex128)

    elif n == 2:
        return np.array(
            [
                [[0, 1], [1, 0]],
                [[0, -1j], [1j, 0]],
                [[1, 0], [0, -1]],
            ], dtype=np.complex128,
        )

    else:
        previous_generators = compute_su_generators(n - 1)
        generators = [
            np.pad(previous_generators[i], (0, 1))
            for i in range(len(previous_generators))
        ]
        for i in range(n - 1):
            t = np.zeros((n, n), dtype=np.complex128)
            t[i, n - 1] = 1.0
            t[n - 1, i] = 1.0
            generators.append(t)
            t2 = np.zeros((n, n), dtype=np.complex128)
            t2[i, n - 1] = -1j
            t2[n - 1, i] = 1j
            generators.append(t2)

        t3 = np.eye(n)
        t3[n - 1, n - 1] = -n + 1
        t3 *= np.sqrt(2 / (n * (n - 1)))
        generators.append(t3)
        return np.array(generators, dtype=np.complex128)


def canonical_unitary(
    unitary: npt.NDArray[np.complex128],
) -> npt.NDArray[np.complex128]:
    """
    Computes a canonical form for the provided unitary.

    If unitary matrices V, W differ only by a global phase, then
    canonical_unitary(V) == canonical_unitary(W).

    Args:
        unitary (npt.NDArray[np.complex128]): A unitary matrix.

    Returns:
        npt.NDArray[np.complex128]: A unitary matrix.

    References:
        https://arxiv.org/abs/2306.05622
    """
    determinant = np.linalg.det(unitary)
    dimension = len(unitary)
    # Compute special unitary
    global_phase = np.angle(determinant) / dimension
    global_phase = global_phase % (2 * np.pi / dimension)
    global_phase_factor = np.exp(-1j * global_phase)
    special_unitary = global_phase_factor * unitary
    # Standardize speical unitary to account for exp(-i2pi/N) differences
    first_row_mags = np.linalg.norm(special_unitary[0, :], ord=2)
    index = np.argmax(first_row_mags)
    std_phase = np.angle(special_unitary[0, index])
    correction_phase = 0 - std_phase
    std_correction = np.exp(1j * correction_phase)
    return std_correction * special_unitary
